"""Plotting bases for specialised plotting."""

from __future__ import annotations

from dataclasses import dataclass

import atlasify
from matplotlib import axis, gridspec, lines
from matplotlib.figure import Figure

from puma.utils import logger, set_xaxis_ticklabels_invisible

atlasify.LINE_SPACING = 1.3  # overwrite the default, which is 1.2


@dataclass
class PlotLineObject:
    """Base data class defining properties of a plot object.

    Parameters
    ----------
    xmin : float, optional
        Minimum value of the x-axis, by default None
    xmax : float, optional
        Maximum value of the x-axis, by default None
    colour : str, optional
        Colour of the object, by default None
    label : str, optional
        Label of object, by default None
    linestyle : str, optional
        Linestyle following numpy style, by default None
    alpha : float, optional
        Value for visibility of the plot lines, by default None
    marker : str, optional
        Marker that is used in the plot. For example an x.
        By default None
    markersize : int, optional
        Size of the marker. By default None
    markeredgewidth : int, optional
        Edge width of the marker. By default None
    is_marker : bool, optional
        Bool, to give info about if this is a marker or a line.
        By default None
    """

    xmin: float = None
    xmax: float = None
    colour: str = None
    label: str = None
    linestyle: str = None
    linewidth: str = None
    alpha: float = None
    marker: str = None
    markersize: int = None
    markeredgewidth: int = None
    is_marker: bool = None


@dataclass
class PlotObject:
    """Data base class defining properties of a plot object.

    Parameters
    ----------
    title : str, optional
        Title of the plot, by default ""
    draw_errors : bool, optional
        Draw statistical uncertainty on the lines, by default True
    xmin : float, optional
        Minimum value of the x-axis, by default None
    xmax : float, optional
        Maximum value of the x-axis, by default None
    ymin : float, optional
        Minimum value of the y-axis, by default None
    ymax : float, optional
        Maximum value of the y-axis, by default None
    ymin_ratio : list, optional
        Set the lower y limit of each of the ratio subplots, by default None.
    ymax_ratio : list, optional
        Set the upper y limit of each of the ratio subplots, by default None.
    y_scale : float, optional
        Scaling up the y axis, e.g. to fit the ATLAS Tag. Applied if ymax not defined,
        by default 1.3
    logx: bool, optional
        Set the log of x-axis, by default False
    logy : bool, optional
        Set log of y-axis of main panel, by default True
    xlabel : str, optional
        Label of the x-axis, by default None
    ylabel : str, optional
        Label of the y-axis, by default None
    ylabel_ratio : list, optional
        List of labels for the y-axis in the ratio plots, by default "Ratio"
    label_fontsize : int, optional
        Used fontsize in label, by default 12
    fontsize : int, optional
        Used fontsize, by default 10
    n_ratio_panels : int, optional
        Amount of ratio panels between 0 and 2, by default 0
    vertical_split: bool
        Set to False if you would like to split the figure horizonally. If set to
        True the figure is split vertically (e.g for pie chart). By default False.
    figsize : (float, float), optional
        Tuple of figure size `(width, height)` in inches, by default (8, 6)
    dpi : int, optional
        DPI used for plotting, by default 400
    transparent : bool, optional
        Specify if the background of the plot should be transparent, by default False
    grid : bool, optional
        Set the grid for the plots.
    leg_fontsize : int, optional
        Fontsize of the legend, by default 10
    leg_loc : str, optional
        Position of the legend in the plot, by default "upper right"
    leg_ncol : int, optional
        Number of legend columns, by default 1
    leg_linestyle_loc : str, optional
        Position of the linestyle legend in the plot, by default "upper center"
    apply_atlas_style : bool, optional
        Apply ATLAS style for matplotlib, by default True
    use_atlas_tag : bool, optional
        Use the ATLAS Tag in the plots, by default True
    atlas_first_tag : str, optional
        First row of the ATLAS tag (i.e. the first row is "ATLAS <atlas_first_tag>"),
        by default "Simulation Internal"
    atlas_second_tag : str, optional
        Second row of the ATLAS tag, by default ""
    atlas_fontsize : float, optional
        Fontsize of ATLAS label, by default 10
    atlas_vertical_offset : float, optional
        Vertical offset of the ATLAS tag, by default 7
    atlas_horizontal_offset : float, optional
        Horizontal offset of the ATLAS tag, by default 8
    atlas_brand : str, optional
        `brand` argument handed to atlasify. If you want to remove it just use an empty
        string or None, by default "ATLAS"
    atlas_tag_outside : bool, optional
        `outside` argument handed to atlasify. Decides if the ATLAS logo is plotted
        outside of the plot (on top), by default False
    atlas_second_tag_distance : float, optional
        Distance between the `atlas_first_tag` and `atlas_second_tag` text in units
        of line spacing, by default 0
    plotting_done : bool
        Bool that indicates if plotting is done. Only then `atlasify()` can be called,
        by default False
    """

    title: str = ""
    draw_errors: bool = True

    xmin: float = None
    xmax: float = None
    ymin: float = None
    ymax: float = None
    ymin_ratio: list = None
    ymax_ratio: list = None
    y_scale: float = 1.3
    logx: bool = False
    logy: bool = True
    xlabel: str = None
    ylabel: str = None
    ylabel_ratio: list = None
    label_fontsize: int = 12
    fontsize: int = 10

    n_ratio_panels: int = 0
    vertical_split: bool = False

    figsize: tuple = None
    dpi: int = 400
    transparent: bool = False

    grid: bool = True

    # legend settings
    leg_fontsize: int = None
    leg_loc: str = "upper right"
    leg_linestyle_loc: str = "upper center"
    leg_ncol: int = 1

    # defining ATLAS style and tags
    apply_atlas_style: bool = True
    use_atlas_tag: bool = True
    atlas_first_tag: str = "Simulation Internal"
    atlas_second_tag: str = None
    atlas_fontsize: int = None
    atlas_vertical_offset: float = 7
    atlas_horizontal_offset: float = 8
    atlas_brand: str = "ATLAS"
    atlas_tag_outside: bool = False
    atlas_second_tag_distance: float = 0

    plotting_done: bool = False

    def __post_init__(self):
        """Check for allowed values.

        Raises
        ------
        ValueError
            If n_ratio_panels not in [0, 1, 2, 3]
        """
        self.__check_figsize()
        allowed_n_ratio_panels = [0, 1, 2, 3]
        if self.n_ratio_panels not in allowed_n_ratio_panels:
            raise ValueError(
                f"{self.n_ratio_panels} not allwed value for `n_ratio_panels`. "
                f"Allowed are {allowed_n_ratio_panels}"
            )
        self.__check_yratio(self.ymin_ratio)
        self.ymin_ratio = (
            [None] * self.n_ratio_panels if self.ymin_ratio is None else self.ymin_ratio
        )
        self.__check_yratio(self.ymax_ratio)
        self.ymax_ratio = (
            [None] * self.n_ratio_panels if self.ymax_ratio is None else self.ymax_ratio
        )

        if self.ylabel_ratio is None:
            self.ylabel_ratio = ["Ratio"] * self.n_ratio_panels
        if len(self.ylabel_ratio) != self.n_ratio_panels:
            raise ValueError(
                f"You passed `ylabel_ratio` of length {len(self.ylabel_ratio)}, "
                f"but `n_ratio_panels` of {self.n_ratio_panels}. "
                f"These should be equal."
            )
        if self.leg_fontsize is None:
            self.leg_fontsize = self.fontsize
        if self.atlas_fontsize is None:
            self.atlas_fontsize = self.fontsize
        if self.apply_atlas_style is False and (
            self.atlas_first_tag is not None or self.atlas_second_tag is not None
        ):
            logger.warning(
                "You specified an ATLAS tag, but `apply_atlas_style` is set to false. "
                "Tag will therefore not be shown on plot."
            )

    def __check_figsize(self):
        """Check `figsize`.

        Raises
        ------
        ValueError
            If shape of `figsize` is not a tuple or list with length 2
        """
        if self.figsize is None:
            return
        if isinstance(self.figsize, list) and len(self.figsize) == 2:
            self.figsize = tuple(self.figsize)
        elif not isinstance(self.figsize, tuple) or len(self.figsize) != 2:
            raise ValueError(
                f"You passed `figsize` as {self.figsize} which is not allowed. "
                "Either a tuple or a list of size 2 is allowed"
            )

    def __check_yratio(self, yratio):
        """Check `yratio`.

        Parameters
        ----------
        yratio : list
            List of min or max limits of ratio plots

        Raises
        ------
        ValueError
            If `yratio` is not a list and it's length
            is not equal to number of ratio panels
        """
        if yratio is None:
            return
        if not isinstance(yratio, (list, tuple)) or len(yratio) != self.n_ratio_panels:
            raise ValueError(
                f"You passed `min/max_yratio` as {yratio} which is not allowed. "
                f"Either a tuple or a list of size {self.n_ratio_panels} is allowed"
            )


class PlotBase(PlotObject):
    """Base class for plotting."""

    def __init__(self, **kwargs) -> None:
        """Initialise class.

        Parameters
        ----------
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`
        """
        super().__init__(**kwargs)
        self.axis_top = None
        self.ratio_axes = []
        self.axis_leg = None
        self.fig = None

    def initialise_figure(self):
        """
        Initialising matplotlib.figure.Figure for different scenarios depending on how
        many ratio panels are requested.
        """
        if self.vertical_split:  # split figure vertically instead of horizonally
            if self.n_ratio_panels >= 1:
                logger.warning(
                    "You set the number of ratio panels to %i but also set the"
                    " vertical splitting to True. Therefore no ratiopanels are"
                    " created.",
                    self.n_ratio_panels,
                )
            self.fig = Figure(figsize=(6, 4.5) if self.figsize is None else self.figsize)
            g_spec = gridspec.GridSpec(1, 11, figure=self.fig)
            self.axis_top = self.fig.add_subplot(g_spec[0, :9])
            self.axis_leg = self.fig.add_subplot(g_spec[0, 9:])

        else:
            # you must use increments of 0.1 for the dimensions
            width = 4.7
            top_height = 2.7 if self.n_ratio_panels else 3.5
            ratio_height = 1.0
            height = top_height + self.n_ratio_panels * ratio_height
            figsize = (width, height) if self.figsize is None else self.figsize
            self.fig = Figure(figsize=figsize, constrained_layout=True)

            if self.n_ratio_panels == 0:
                self.axis_top = self.fig.gca()
            elif self.n_ratio_panels > 0:
                g_spec_height = (top_height + ratio_height * self.n_ratio_panels) * 10
                g_spec = gridspec.GridSpec(int(g_spec_height), 1, figure=self.fig)
                self.axis_top = self.fig.add_subplot(g_spec[: int(top_height * 10), 0])
                set_xaxis_ticklabels_invisible(self.axis_top)
                for i in range(1, self.n_ratio_panels + 1):
                    start = int((top_height + ratio_height * (i - 1)) * 10)
                    stop = int(start + ratio_height * 10)
                    sub_axis = self.fig.add_subplot(g_spec[start:stop, 0], sharex=self.axis_top)
                    if i < self.n_ratio_panels:
                        set_xaxis_ticklabels_invisible(sub_axis)
                    self.ratio_axes.append(sub_axis)

        if self.grid:
            self.axis_top.grid(lw=0.3)
            for ratio_axis in self.ratio_axes:
                ratio_axis.grid(lw=0.3)

    def draw_vlines(
        self,
        xs: list,
        labels: list | None = None,
        ys: list | None = None,
        same_height: bool = False,
        colour: str = "#000000",
        linestyle: str = "dashed",
        fontsize: int = 10,
    ):
        """Drawing working points in plot.

        Parameters
        ----------
        xs : list
            List of working points x values to draw
        labels : list, optional
            List with labels for the vertical lines. Must be the same
            order as the xs. If None, the xvalues * 100 will be
            used as labels. By default None
        ys : list, optional
            List with the y height of the vertical lines in percent of the
            upper plot (0 is bottom, 1 is top). Must be the same
            order as the xs and the labels. By default None
        same_height : bool, optional
            Working point lines on same height, by default False
        colour : str, optional
            Colour of the vertical line, by default "#000000" (black)
        linestyle : str, optional
            Linestyle of the vertical line, by default "dashed"
        fontsize : int, optional
            Fontsize of the vertical line text. By default 10.
        """
        for i, vline_x in enumerate(xs):
            # Set y-point of the WP lines/text
            ytext = (0.65 if same_height else 0.65 - i * 0.07) if ys is None else ys[i]

            self.axis_top.axvline(
                x=vline_x,
                ymax=ytext,
                color=colour,
                linestyle=linestyle,
                linewidth=1.0,
            )

            # Set the number above the line
            self.axis_top.text(
                x=vline_x - 0.005,
                y=ytext + 0.005,
                s=labels[i] if labels else None,
                transform=self.axis_top.get_xaxis_text1_transform(0)[0],
                fontsize=fontsize,
            )

            for ratio_axis in self.ratio_axes:
                ratio_axis.axvline(x=vline_x, color=colour, linestyle=linestyle, linewidth=1.0)

    def set_title(self, title: str | None = None, **kwargs):
        """Set title of top panel.

        Parameters
        ----------
        title : str, optional
            Title of top panel, if None using the value form the class variables,
            by default None
        **kwargs : kwargs
            Keyword arguments passed to `matplotlib.axes.Axes.set_title()`
        """
        self.axis_top.set_title(self.title if title is None else title, **kwargs)

    def set_log(self):
        """Set log scale of the axes.

        For the y-axis, only the main panel is set. For the x-axes also set the
        ratio panels.
        """
        if self.logx:
            self.axis_top.set_xscale("log")
            for ratio_axis in self.ratio_axes:
                ratio_axis.set_xscale("log")

        if self.logy:
            self.axis_top.set_yscale("log")
            ymin, ymax = self.axis_top.get_ylim()
            self.y_scale = ymin * ((ymax / ymin) ** self.y_scale) / ymax

    def set_y_lim(self):
        """Set limits of y-axis."""
        ymin, ymax = self.axis_top.get_ylim()
        self.axis_top.set_ylim(
            ymin if self.ymin is None else self.ymin,
            ymin + (ymax - ymin) * self.y_scale if self.ymax is None else self.ymax,
        )

        for i, ratio_axis in enumerate(self.ratio_axes):
            if self.ymin_ratio[i] is not None or self.ymax_ratio[i] is not None:
                ymin, ymax = ratio_axis.get_ylim()
                ymin = self.ymin_ratio[i] if self.ymin_ratio[i] is not None else ymin
                ymax = self.ymax_ratio[i] if self.ymax_ratio[i] is not None else ymax
                ratio_axis.set_ylim(bottom=ymin, top=ymax)

    def set_ylabel(self, ax_mpl, label: str | None = None, align: str | None = "right", **kwargs):
        """Set y-axis label.

        Parameters
        ----------
        ax_mpl : matplotlib.axes.Axes
            matplotlib axis object
        label : str, optional
            x-axis label, by default None
        align : str, optional
            Alignment of y-axis label, by default "right"
        **kwargs, kwargs
            Keyword arguments passed to `matplotlib.axes.Axes.set_ylabel()`
        """
        label_options = {"fontsize": self.label_fontsize}
        if align:
            label_options["horizontalalignment"] = align
            if align == "right":
                label_options["y"] = 1
            elif align == "left":
                label_options["y"] = 0

        ax_mpl.set_ylabel(
            self.ylabel if label is None else label,
            **label_options,
            **kwargs,
        )
        self.fig.align_labels()

    def set_xlabel(self, label: str | None = None, **kwargs):
        """Set x-axis label.

        Parameters
        ----------
        label : str, optional
            x-axis label, by default None
        **kwargs : kwargs
            Keyword arguments passed to `matplotlib.axes.Axes.set_xlabel()`
        """
        xlabel_args = {
            "xlabel": self.xlabel if label is None else label,
            "horizontalalignment": "right",
            "x": 1.0,
            "fontsize": self.label_fontsize,
        }
        if self.n_ratio_panels == 0:
            self.axis_top.set_xlabel(**xlabel_args, **kwargs)
        if self.n_ratio_panels > 0:
            self.ratio_axes[-1].set_xlabel(**xlabel_args, **kwargs)

    def set_tick_params(self, labelsize: int | None = None, **kwargs):
        """Set x-axis label.

        Parameters
        ----------
        labelsize : int, optional
            Label size of x- and y- axis ticks, by default None.
            If None then using global fontsize
        **kwargs : kwargs
            Keyword arguments passed to `matplotlib.axes.Axes.set_xlabel()`
        """
        labelsize = self.fontsize if labelsize is None else labelsize
        self.axis_top.tick_params(axis="y", labelsize=labelsize, **kwargs)
        if self.n_ratio_panels == 0:
            self.axis_top.tick_params(axis="x", labelsize=labelsize, **kwargs)
        for i, ratio_axis in enumerate(self.ratio_axes):
            ratio_axis.tick_params(axis="y", labelsize=labelsize, **kwargs)
            if i == self.n_ratio_panels - 1:
                ratio_axis.tick_params(axis="x", labelsize=labelsize, **kwargs)

    def set_xlim(self, xmin: float | None = None, xmax: float | None = None, **kwargs):
        """Set limits of x-axis.

        Parameters
        ----------
        xmin : float, optional
            Min of x-axis, by default None
        xmax : float, optional
            Max of x-axis, by default None
        **kwargs : kwargs
            Keyword arguments passed to `matplotlib.axes.Axes.set_xlim()`
        """
        self.axis_top.set_xlim(
            self.xmin if xmin is None else xmin,
            self.xmax if xmax is None else xmax,
            **kwargs,
        )

    def savefig(
        self,
        plot_name: str,
        transparent: bool | None = None,
        dpi: int | None = None,
        **kwargs,
    ):
        """Save plot to disk.

        Parameters
        ----------
        plot_name : str
            File name of the plot
        transparent : bool, optional
            Specify if plot background is transparent, by default False
        dpi : int, optional
            DPI for plotting, by default 400
        **kwargs : kwargs
            Keyword arguments passed to `matplotlib.figure.Figure.savefig()`
        """
        logger.debug("Saving plot to %s", plot_name)
        self.fig.savefig(
            plot_name,
            transparent=self.transparent if transparent is None else transparent,
            dpi=self.dpi if dpi is None else dpi,
            bbox_inches="tight",
            pad_inches=0.04,
            **kwargs,
        )

    def atlasify(self, force: bool = False):
        """Apply ATLAS style to all axes using the atlasify package.

        Parameters
        ----------
        force : bool, optional
            Force ATLAS style also if class variable is False, by default False
        """
        if self.plotting_done is False and force is False:
            logger.warning(
                "`atlasify()` has to be called after plotting --> "
                "ATLAS style will not be adapted. If you want to do it anyway, "
                "you can use `force`."
            )
            return

        if self.apply_atlas_style or force:
            logger.debug("Initialise ATLAS style using atlasify.")
            if self.use_atlas_tag is True:
                atlasify.atlasify(
                    atlas=self.atlas_first_tag,
                    subtext=self.atlas_second_tag,
                    axes=self.axis_top,
                    font_size=self.atlas_fontsize,
                    label_font_size=self.atlas_fontsize,
                    sub_font_size=self.atlas_fontsize,
                    offset=self.atlas_vertical_offset,
                    indent=self.atlas_horizontal_offset,
                    enlarge=1,
                    brand="" if self.atlas_brand is None else self.atlas_brand,
                    outside=self.atlas_tag_outside,
                    subtext_distance=self.atlas_second_tag_distance,
                )
            else:
                atlasify.atlasify(atlas=False, axes=self.axis_top, enlarge=1)
            for ratio_axis in self.ratio_axes:
                atlasify.atlasify(atlas=False, axes=ratio_axis, enlarge=1)
            if self.vertical_split:
                atlasify.atlasify(atlas=False, axes=self.axis_leg, enlarge=1)
            if force:
                if self.apply_atlas_style is False:
                    logger.warning(
                        "Initialising ATLAS style even though `apply_atlas_style` is "
                        " set to False."
                    )
                if self.plotting_done is False:
                    logger.warning(
                        "Initialising ATLAS style even though `plotting_done` is set to" " False."
                    )

    def make_legend(self, handles: list, ax_mpl: axis, labels: list | None = None, **kwargs):
        """Drawing legend on axis.

        Parameters
        ----------
        handles :  list
            List of matplotlib.lines.Line2D object returned when plotting
        ax_mpl : matplotlib.axis.Axes
            `matplotlib.axis.Axes` object where the legend should be plotted
        labels : list, optional
            Plot labels. If None, the labels are extracted from the `handles`.
            By default None
        **kwargs : kwargs
            Keyword arguments which can be passed to matplotlib axis
        """
        if labels is None:
            # remove the handles which have label=None
            handles = [handle for handle in handles if handle.get_label() is not None]
        ax_mpl.add_artist(
            ax_mpl.legend(
                handles=handles,
                labels=([handle.get_label() for handle in handles] if labels is None else labels),
                loc=self.leg_loc,
                fontsize=self.leg_fontsize,
                ncol=self.leg_ncol,
                **kwargs,
            )
        )

    def make_linestyle_legend(
        self,
        linestyles: list,
        labels: list,
        loc: str | None = None,
        bbox_to_anchor: tuple | None = None,
        axis_for_legend=None,
    ):
        """Create a legend to indicate what different linestyles correspond to.

        Parameters
        ----------
        linestyles : list
            List of the linestyles to draw in the legend
        labels : list
            List of the corresponding labels. Has to be in the same order as the
            linestyles
        loc : str, optional
            Location of the legend (matplotlib supported locations), by default None
        bbox_to_anchor : tuple, optional
            Allows to specify the precise position of this legend. Either a 2-tuple
            (x, y) or a 4-tuple (x, y, width, height), by default None
        axis_for_legend : matplotlib.Axes.axis, optional
            Axis on which to draw the legend, by default None
        """
        if axis_for_legend is None:
            axis_for_legend = self.axis_top

        lines_list = []
        for linestyle, label in zip(linestyles, labels):
            lines_list.append(
                lines.Line2D(
                    [],
                    [],
                    color="k",
                    label=label,
                    linestyle=linestyle,
                )
            )

        linestyle_legend = axis_for_legend.legend(
            handles=lines_list,
            labels=[handle.get_label() for handle in lines_list],
            loc=loc if loc is not None else self.leg_linestyle_loc,
            fontsize=self.leg_fontsize,
            bbox_to_anchor=bbox_to_anchor,
            frameon=False,
        )
        axis_for_legend.add_artist(linestyle_legend)

    def set_ratio_label(self, ratio_panel: int, label: str):
        """Associate the rejection class to a ratio panel.

        Parameters
        ----------
        ratio_panel : int
            Index of the ratio panel to label.
        label : str
            y-axis label of the ratio panel.

        Raises
        ------
        ValueError
            If the requested ratio panel does not exist.
        """
        if ratio_panel > self.n_ratio_panels:
            raise ValueError(f"Plot has {self.n_ratio_panels} ratio panels, not {ratio_panel}")
        self.ylabel_ratio[ratio_panel - 1] = label



from __future__ import annotations

import copy
from pathlib import Path

from ftag import Label
from ftag.hdf5 import H5Reader

from puma.utils import logger


def combine_suffixes(suffixes):
    """Combines a list of suffixes into a single suffix."""
    clean_suffixes = [s for s in suffixes if s not in {"", None}]
    if len(clean_suffixes) == 0:
        return None
    return "_".join([s for s in suffixes if s not in {"", None}])


def get_include_exclude_str(include_taggers, all_taggers):
    """Generates the name of the plot, based on the included taggers."""
    if len(include_taggers) == len(all_taggers):
        return ""

    return "taggers_" + "_".join([t.yaml_name for t in include_taggers.values()])


def get_included_taggers(results, plot_config):
    """Converts 'include_taggers' or 'exclude_taggers' into a list of the taggers
    to include.
    """
    all_taggers = results.taggers
    all_tagger_names = [t.yaml_name for t in all_taggers.values()]
    if "include_taggers" not in plot_config and "exclude_taggers" not in plot_config:
        include_taggers = results.taggers
    elif include_taggers := plot_config.get("include_taggers", None):
        assert all(
            t in all_tagger_names for t in include_taggers
        ), f"Not all taggers are in the results: {include_taggers}"

        include_taggers = {
            t: v for t, v in results.taggers.items() if v.yaml_name in include_taggers
        }

    elif exclude_taggers := plot_config.get("exclude_taggers", None):
        assert all(t in all_tagger_names for t in exclude_taggers)
        include_taggers = {
            t: v for t, v in results.taggers.items() if v.yaml_name not in exclude_taggers
        }

    if len(include_taggers) == 0:
        raise ValueError(
            "No taggers included in plot, check that 'exclude_taggers' doesn't exclude "
            "all taggers, or that atleast 1 tagger is defined in 'include_taggers'"
        )
    logger.debug("Include taggers: %s", include_taggers)

    # Set which tagger to use as a reference, if no reference is set, use the first
    #  tagger.This is only needed for plots with a ratio, but still...
    if not any(t.reference for t in include_taggers.values()):
        if reference := plot_config.get("reference", None):
            if reference not in [t.yaml_name for t in include_taggers.values()]:
                raise ValueError(
                    f"Reference {reference} not in included taggers" f" {include_taggers.keys()}"
                )
            reference = str(next(t for t in include_taggers.values() if t.yaml_name == reference))
            # Create a copy, and set it as reference, this is the easiest way of doing
            #  this but might be a bit slow
        else:
            reference = next(iter(include_taggers.keys()))
            logger.info("No reference set for plot, using " + reference + " as reference")

        include_taggers[reference] = copy.deepcopy(include_taggers[reference])
        include_taggers[reference].reference = True

    return (
        include_taggers,
        all_taggers,
        get_include_exclude_str(include_taggers, all_taggers),
    )


def get_tagger_name(name: str, sample_path: Path, key: str, flavours: list[Label]):
    """Attempts to return the name of the tagger if it is not specified in the config
    file by looking at available variable names, and the key of the tagger in the config.

    Parameters
    ----------
    name : str
        The name of the tagger, if specified in the config file
    sample_path : Path
        The path to the sample file
    key : str
        The key of the tagger in the config file, if multiple taggers are found in the
        sample path, this is used to select the correct tagger
    flavours : list[Label]
        The flavours of the tagger, used to identify the correct tagger. A 'valid'
        tagger in the file is one where tagger_p{flav} exists for all flavours defined
        in this list

    Returns
    -------
    str
        The name of the tagger to use
    """
    if name:
        return name

    reader = H5Reader(sample_path)
    jet_vars = reader.dtypes()["jets"].names
    req_keys = [f"_p{flav.name[:-4]}" for flav in flavours]

    potential_taggers = {}

    # Identify potential taggers
    for var in jet_vars:
        for suffix in req_keys:
            if var.endswith(suffix):
                base_name = var.rsplit(suffix, 1)[0]
                if base_name in potential_taggers:
                    potential_taggers[base_name].append(suffix)
                else:
                    potential_taggers[base_name] = [suffix]

    # Check if any base name has all three suffixes
    valid_taggers = [
        base for base, suffixes in potential_taggers.items() if set(suffixes) == set(req_keys)
    ]

    if len(valid_taggers) == 0:
        raise ValueError("No valid tagger found.")
    if len(valid_taggers) > 1:
        if key in valid_taggers:
            return key
        raise ValueError(
            f"Multiple valid taggers found: {', '.join(valid_taggers)} in file "
            f"{sample_path}, please specify the tagger name in the config file"
        )
    return valid_taggers[0]



"""Support functions for SV mass calculation."""

from __future__ import annotations

import numpy as np


def calculate_vertex_mass(pt, eta, phi, vtx_idx, particle_mass=0.13957):
    """
    Calculate the invariant mass of secondary vertices from the track 4-momenta,
    assuming a pion mass hypothesis.

    Parameters
    ----------
    pt: np.ndarray
        Array of shape (n_jets, n_tracks) containing the track transverse momenta (in GeV).
    eta: np.ndarray
        Array of shape (n_jets, n_tracks) containing the track pseudorapidities.
    phi: np.ndarray
        Array of shape (n_jets, n_tracks) containing the track azimuthal angles.
    vtx_idx: np.ndarray
        Array of shape (n_jets, n_tracks) containing the vertex indices for each track.
    particle_mass: float, optional
        Mass hypothesis for each particle in GeV. Default is the pion mass.

    Returns
    -------
    mass: np.ndarray
        Array of shape (n_jets, n_tracks) containing the invariant vertex mass for each track.
    """
    n_jets = pt.shape[0]
    sv_masses = np.full(pt.shape, -1.0, dtype=float)

    for i in range(n_jets):
        vtx_idx_i = vtx_idx[i]
        pt_i = pt[i]
        eta_i = eta[i]
        phi_i = phi[i]

        unique_idx = np.unique(vtx_idx_i)
        unique_idx = unique_idx[unique_idx >= 0]  # remove tracks with negative indices
        vertices = np.tile(vtx_idx_i, (unique_idx.size, 1))
        comparison_ids = np.tile(unique_idx, (vtx_idx_i.size, 1)).T
        vertices = (vertices == comparison_ids).astype(int)

        pt_i = pt_i * vertices
        eta_i = eta_i * vertices
        phi_i = phi_i * vertices
        m_i = particle_mass * vertices

        px_i = pt_i * np.cos(phi_i)
        py_i = pt_i * np.sin(phi_i)
        pz_i = pt_i * np.sinh(eta_i)
        e_i = np.sqrt(px_i**2 + py_i**2 + pz_i**2 + m_i**2)

        px = np.nansum(px_i, axis=1)
        py = np.nansum(py_i, axis=1)
        pz = np.nansum(pz_i, axis=1)
        e = np.nansum(e_i, axis=1)

        m = np.sqrt(e**2 - px**2 - py**2 - pz**2)
        sv_masses[i] = np.sum(np.tile(m, (vtx_idx_i.size, 1)).T * vertices, axis=0)

    return sv_masses



"""Results module for high level API."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
from ftag import Cuts, Flavours, Label
from ftag.hdf5 import H5Reader
from matplotlib.figure import Figure

from puma import (
    Histogram,
    HistogramPlot,
    Line2D,
    Line2DPlot,
    Roc,
    RocPlot,
    VarVsEff,
    VarVsEffPlot,
    fraction_scan,
)
from puma.hlplots.tagger import Tagger
from puma.hlplots.yutils import combine_suffixes
from puma.metrics import calc_eff, calc_rej
from puma.utils import get_good_colours, get_good_linestyles, logger


@dataclass
class Results:
    """Store information about several taggers and plot results."""

    signal: Label | str
    sample: str
    backgrounds: list = field(init=False)
    all_flavours: list[Label] | None = None
    atlas_first_tag: str = "Simulation Internal"
    atlas_second_tag: str = None
    atlas_third_tag: str = None
    taggers: dict = field(default_factory=dict)
    perf_vars: str | tuple | list = "pt"
    output_dir: str | Path = "."
    extension: str = "png"
    global_cuts: Cuts | list | None = None
    num_jets: int | None = None
    remove_nan: bool = False
    label_var: str = "HadronConeExclTruthLabelID"

    def __post_init__(self):
        if self.all_flavours:
            self.all_flavours = [Flavours[f] for f in self.all_flavours]
        self.set_signal(self.signal)
        if isinstance(self.output_dir, str):
            self.output_dir = Path(self.output_dir)
        if isinstance(self.perf_vars, str):
            self.perf_vars = [self.perf_vars]
        if self.atlas_second_tag is not None and self.atlas_third_tag is not None:
            self.atlas_second_tag = f"{self.atlas_second_tag}\n{self.atlas_third_tag}"

        self.plot_funcs = {
            "probs": self.plot_probs,
            "disc": self.plot_discs,
            "roc": self.plot_rocs,
            "peff": self.plot_var_perf,
            "scan": self.plot_fraction_scans,
        }
        self.saved_plots = []

    def set_signal(self, signal: Label):
        """Set the signal flavour and define background flavours.

        Parameters
        ----------
        signal : Flavour
            Flavour which is the signal

        Raises
        ------
        ValueError
            If the signal class is not supported
        """
        if isinstance(signal, str):
            signal = Flavours[signal]
        self.signal = signal
        # If we have a list of al flavours, then the background is always [all - signal]
        if self.all_flavours:
            self.backgrounds = [f for f in self.all_flavours if f != self.signal]
            return
        if self.signal == Flavours.bjets:
            self.backgrounds = [Flavours.cjets, Flavours.ujets]
        elif self.signal == Flavours.cjets:
            self.backgrounds = [Flavours.bjets, Flavours.ujets]
        elif self.signal == Flavours.hbb:
            self.backgrounds = [Flavours.hcc, Flavours.top, Flavours.qcd]
        elif self.signal == Flavours.hcc:
            self.backgrounds = [Flavours.hbb, Flavours.top, Flavours.qcd]
        else:
            raise ValueError(f"Unsupported signal class {self.signal}.")

    @property
    def sig_str(self):
        """Returns the name of the signal as a string.

        Returns
        -------
        str
            Name of the signal as string
        """
        suffix = "jets"
        sig = str(self.signal)
        if sig.endswith(suffix):
            sig = sig[: -len(suffix)]
        return f"{sig}"

    @property
    def flavours(self):
        """Return a list of all flavours.

        Returns
        -------
        list
            List of all flavours
        """
        return [*self.backgrounds, self.signal]

    def add(self, tagger: Tagger):
        """Add tagger to class.

        Parameters
        ----------
        tagger : puma.hlplots.Tagger
            Instance of the puma.hlplots.Tagger class, containing tagger information.

        Raises
        ------
        KeyError
            if model name duplicated
        """
        if str(tagger) in self.taggers:
            raise KeyError(f"{tagger} was already added.")
        if not tagger.colour:
            good_colours = get_good_colours()
            current_colours = [t.colour for t in self.taggers.values()]
            tagger.colour = next(c for c in good_colours if c not in current_colours)
        if tagger.output_flavours is None:
            tagger.output_flavours = self.flavours
        self.taggers[str(tagger)] = tagger

    def load(self):
        """Iterates all taggers, and loads data if it hasn't already been loaded."""
        req_load = [tagger for tagger in self.taggers.values() if tagger.scores is None]
        tagger_paths = list({tagger.sample_path for tagger in req_load})
        for tp in tagger_paths:
            tp_taggers = [tagger for tagger in req_load if tagger.sample_path == tp]
            self.load_taggers_from_file(
                tp_taggers,
                tp,
                cuts=self.global_cuts,
                num_jets=self.num_jets,
                label_var=self.label_var,
            )

    def load_taggers_from_file(  # pylint: disable=R0913
        self,
        taggers: list[Tagger],
        file_path: Path | str,
        key="jets",
        label_var="HadronConeExclTruthLabelID",
        cuts: Cuts | list | None = None,
        num_jets: int | None = None,
        perf_vars: dict | None = None,
    ):
        """Load one or more taggers from a common file. Adds the tagger to this results class
        if it is not already present.

        Parameters
        ----------
        taggers : list[Tagger]
            List of taggers to add
        file_path : str | Path
            Path to file
        key : str, optional
            Key in file, by default 'jets'
        label_var : str
            Label variable to use, by default 'HadronConeExclTruthLabelID'
        cuts : Cuts | list, optional
            Cuts to apply, by default None
        num_jets : int, optional
            Number of jets to load from the file, by default all jets
        perf_vars : dict, optional
            Override the performance variables to use, by default None
        """

        def check_nan(data: np.ndarray) -> np.ndarray:
            """
            Filter out NaN values from loaded data.

            Parameters
            ----------
            data : ndarray
                Data to filter
            """
            mask = np.ones(len(data), dtype=bool)
            for name in data.dtype.names:
                mask = np.logical_and(mask, ~np.isnan(data[name]))
            if np.sum(~mask) > 0:
                if self.remove_nan:
                    logger.warning(
                        f"{np.sum(~mask)} NaN values found in loaded data. Removing" " them."
                    )
                    return data[mask]
                raise ValueError(f"{np.sum(~mask)} NaN values found in loaded data.")
            return data

        # set tagger output nodes
        for tagger in taggers:
            if tagger not in self.taggers.values():
                self.add(tagger)
            tagger.output_flavours = self.flavours
            if "ftau" not in tagger.fxs and Flavours.taujets in tagger.output_flavours:
                tagger.output_flavours.remove(Flavours.taujets)

        # get a list of all variables to be loaded from the file
        if not isinstance(cuts, Cuts):
            cuts = Cuts.empty() if cuts is None else Cuts.from_list(cuts)
        var_list = sum([tagger.variables for tagger in taggers], [label_var])
        var_list += cuts.variables
        var_list += sum([t.cuts.variables for t in taggers if t.cuts is not None], [])
        var_list = list(set(var_list + self.perf_vars))

        # load data
        reader = H5Reader(file_path, precision="full")
        data = reader.load({key: var_list}, num_jets)[key]

        # check for nan values
        data = check_nan(data)
        # apply common cuts
        if cuts:
            idx, data = cuts(data)
            if perf_vars is not None:
                for perf_var_name, perf_var_array in perf_vars.items():
                    perf_vars[perf_var_name] = perf_var_array[idx]

        # for each tagger
        for tagger in taggers:
            sel_data = data
            sel_perf_vars = perf_vars

            # apply tagger specific cuts
            if tagger.cuts:
                idx, sel_data = tagger.cuts(data)

            # attach data to tagger objects
            tagger.extract_tagger_scores(sel_data, source_type="structured_array")
            tagger.labels = np.array(sel_data[label_var], dtype=[(label_var, "i4")])
            if perf_vars is None:
                tagger.perf_vars = {}
                for perf_var in self.perf_vars:
                    if any(x in perf_var for x in ["pt", "mass"]):
                        tagger.perf_vars[perf_var] = sel_data[perf_var] * 0.001
                    else:
                        tagger.perf_vars[perf_var] = sel_data[perf_var]
            else:
                tagger.perf_vars = sel_perf_vars

    def __getitem__(self, tagger_name: str):
        """Retrieve Tagger object.

        Parameters
        ----------
        tagger_name : str
            Name of model

        Returns
        -------
        Tagger
            Instance of the puma.hlplots.Tagger class, containing tagger information.
        """
        return self.taggers[tagger_name]

    def save(
        self,
        plot: Figure,
        plot_type: str,
        base: str | None = None,
        suffix: str | None = None,
    ):
        """Get the output file path.

        Parameters
        ----------
        plot : Figure
            Matplotlib figure to save.
        plot_type : str
            Plots of the same type are saved in the same directory.
        base_fname : str
            Base filename, modified by this function.
        suffix : str, optional
            Suffix to add to the filename, by default None
        """
        tag_str = f"{self.sig_str}tag"
        out_dir = self.output_dir / tag_str / plot_type
        out_dir.mkdir(parents=True, exist_ok=True)
        if not base:
            base = plot_type
        fname = f"{self.sample}_{tag_str}"
        fname += f"_{base}"
        if suffix:
            fname += f"_{suffix}"
        fpath = out_dir / f"{fname}.{self.extension}"
        plot.savefig(fpath)
        self.saved_plots.append(fpath)

    def plot_probs(
        self,
        suffix: str | None = None,
        **kwargs,
    ):
        """Plot probability distributions.

        Parameters
        ----------
        suffix : str, optional
            Suffix to add to output file name, by default None
        **kwargs : kwargs
            key word arguments for `puma.HistogramPlot`
        """
        # Get good linestyles for plotting
        line_styles = get_good_linestyles()

        # Get a list of all flavours which are present
        flavours = [*self.backgrounds, self.signal]

        # Init a default kwargs dict for the HistogramPlot
        histo_kwargs = {
            "ylabel": "Normalised number of jets",
            "figsize": (7.0, 4.5),
            "n_ratio_panels": 1,
            "atlas_first_tag": self.atlas_first_tag,
            "atlas_second_tag": self.atlas_second_tag,
        }

        # If kwargs are given, update the histo_kwargs dict
        if kwargs is not None:
            histo_kwargs.update(kwargs)

        # group by output probability
        for flav_prob in flavours:
            # Create a new histogram plot
            hist = HistogramPlot(
                xlabel=flav_prob.px,
                **histo_kwargs,
            )

            # Init a new list for the tagger labels
            tagger_labels = []

            # Loop over the taggers
            for counter, tagger in enumerate(self.taggers.values()):
                # Append labels if existing else the name
                tagger_labels.append(tagger.label if tagger.label else tagger.name)

                # Add the probability output of the given tagger for each flavour
                for flav_class in flavours:
                    hist.add(
                        Histogram(
                            tagger.probs(flav_prob, flav_class),
                            ratio_group=flav_class,
                            label=flav_class.label if counter == 0 else None,
                            colour=flav_class.colour,
                            linestyle=line_styles[counter],
                        ),
                        reference=tagger.reference,
                    )

            # Finalise the plot and draw it
            hist.draw()
            hist.make_linestyle_legend(
                linestyles=line_styles,
                labels=tagger_labels,
                bbox_to_anchor=(0.55, 1),
            )
            self.save(hist, "prob", flav_prob.px, suffix)

        # group by flavour
        for flav_class in flavours:
            # Create a new histogram plot
            hist = HistogramPlot(
                xlabel=flav_class.label,
                **histo_kwargs,
            )

            # Init a new list for the tagger labels
            tagger_labels = []

            # Loop over the taggers
            for counter, tagger in enumerate(self.taggers.values()):
                # Append labels if existing else the name
                tagger_labels.append(tagger.label if tagger.label else tagger.name)

                # Add the probability output of the given tagger for each flavour
                for flav_prob in flavours:
                    hist.add(
                        Histogram(
                            tagger.probs(flav_prob, flav_class),
                            ratio_group=flav_prob,
                            label=flav_prob.px if counter == 0 else None,
                            colour=flav_prob.colour,
                            linestyle=line_styles[counter],
                        ),
                        reference=tagger.reference,
                    )

            # Finalise the plot and draw it
            hist.draw()
            hist.make_linestyle_legend(
                linestyles=line_styles,
                labels=tagger_labels,
                bbox_to_anchor=(0.55, 1),
            )
            self.save(hist, "prob", flav_class, suffix)

    def plot_discs(
        self,
        suffix: str | None = None,
        exclude_tagger: list | None = None,
        xlabel: str | None = None,
        wp_vlines: list | None = None,
        **kwargs,
    ):
        """Plot discriminant distributions.

        Parameters
        ----------
        suffix : str, optional
            Suffix to add to output file name, by default None
        exclude_tagger : list, optional
            List of taggers to be excluded from this plot, by default None
        xlabel : str, optional
            x-axis label, by default "$D_{b}$"
        wp_vlines : list, optional
            List of WPs to draw vertical lines at, by default None
        **kwargs : kwargs
            key word arguments for `puma.HistogramPlot`
        """
        if xlabel is None:
            xlabel = rf"$D_{{{self.signal.name.rstrip('jets')}}}$"
        if wp_vlines is None:
            wp_vlines = []

        # Get good linestyles for plotting
        line_styles = get_good_linestyles()

        # Init histo_kwargs
        histo_kwargs = {
            "n_ratio_panels": 0,
            "xlabel": xlabel,
            "ylabel": "Normalised number of jets",
            "figsize": (7.0, 4.5),
            "atlas_first_tag": self.atlas_first_tag,
            "atlas_second_tag": self.atlas_second_tag,
        }

        # Check if kwargs are given and update the histo_kwargs accordingly
        if kwargs is not None:
            histo_kwargs.update(kwargs)

        # Create a new histogram plot
        hist = HistogramPlot(**histo_kwargs)

        # Init a tagger label list
        tagger_labels = []

        # Loop over the defined taggers
        for counter, tagger in enumerate(self.taggers.values()):
            # Check if the tagger is excluded and skip it if so
            if exclude_tagger is not None and tagger.name in exclude_tagger:
                continue

            # Get the discriminant values from the tagger for the given signal
            discs = tagger.discriminant(self.signal)

            # Get lists for the working put cuts and labels
            wp_cuts, wp_labels = [], []

            # get working point cuts and labels and append them
            for wp in wp_vlines:
                cut = np.percentile(discs[tagger.is_flav(self.signal)], 100 - wp)
                label = None if counter > 0 else f"{wp}%"
                wp_cuts.append(cut)
                wp_labels.append(label)

            # Draw the vertical lines for the working points
            hist.draw_vlines(wp_cuts, labels=wp_labels, linestyle=line_styles[counter])

            # Loop over the flavours and add the disc values for each flavour
            for flav in self.flavours:
                hist.add(
                    Histogram(
                        discs[tagger.is_flav(flav)],
                        ratio_group=flav,
                        label=flav.label if counter == 0 else None,
                        colour=flav.colour,
                        linestyle=line_styles[counter],
                    ),
                    reference=tagger.reference,
                )

            # Add the tagger label or name to the label list
            tagger_labels.append(tagger.label if tagger.label else tagger.name)

        # Finalise the plot and draw it
        hist.draw()
        hist.make_linestyle_legend(
            linestyles=line_styles,
            labels=tagger_labels,
            bbox_to_anchor=(0.55, 1),
        )
        self.save(hist, "disc", suffix=suffix)

    def plot_rocs(
        self,
        x_range: tuple[float, float] | None = (0.5, 1.0),
        resolution: int = 50,
        suffix: str | None = None,
        **kwargs,
    ):
        """Plots rocs.

        Parameters
        ----------
        x_range : tuple, optional
            x-axis range, by default None
        resolution : int, optional
            number of points to use for the x-axis, by default 100
        suffix : str, optional
            suffix to add to output file name, by default None
        kwargs: dict, optional
            key word arguments being passed to `RocPlot`
        """
        # Linspace the signal efficiencies
        sig_effs = np.linspace(*x_range, resolution)

        # Init a default kwargs dict for roc plots
        roc_kwargs = {
            "n_ratio_panels": len(self.backgrounds),
            "ylabel": "Background rejection",
            "xlabel": self.signal.eff_str,
            "atlas_first_tag": self.atlas_first_tag,
            "atlas_second_tag": self.atlas_second_tag,
            "y_scale": 1.3,
            "ymin": 1,
        }

        # If kwargs are given, update the default kwargs accordingly
        if kwargs is not None:
            kwargs.update(roc_kwargs)

        # Init a new ROC plot using the given/default kwargs
        roc = RocPlot(**kwargs)

        # Iterate over the taggers
        for tagger in self.taggers.values():
            # Get the disc values for the given tagger
            discs = tagger.discriminant(self.signal)

            # Loop over all backgrouns
            for background in self.backgrounds:
                # Calculate rejection for the given background
                rej = calc_rej(
                    discs[tagger.is_flav(self.signal)],
                    discs[tagger.is_flav(background)],
                    sig_effs,
                    smooth=True,
                )

                # Add the rejection curve to the ROC plot
                roc.add_roc(
                    Roc(
                        sig_effs,
                        rej,
                        n_test=tagger.n_jets(background),
                        rej_class=background,
                        signal_class=self.signal,
                        label=tagger.label,
                        colour=tagger.colour,
                    ),
                    reference=tagger.reference,
                )

        # setting which flavour rejection ratio is drawn in which ratio panel
        for counter, background in enumerate(self.backgrounds):
            roc.set_ratio_class(counter + 1, background)

        # Finalise the plot and draw it
        roc.draw()
        self.save(roc, "roc", suffix=suffix)

    def plot_var_perf(  # pylint: disable=too-many-locals
        self,
        suffix: str | None = None,
        xlabel: str = r"$p_{T}$ [GeV]",
        perf_var: str = "pt",
        h_line: float | None = None,
        working_point: float | None = None,
        disc_cut: float | None = None,
        fixed_rejections: dict[Label, float] | None = None,
        **kwargs,
    ):
        """Variable vs efficiency/rejection plot.

        You can choose between different modes: "sig_eff", "bkg_eff", "sig_rej",
        "bkg_rej"

        Parameters
        ----------
        suffix : str, optional
            suffix to add to output file name, by default None
        xlabel : regexp, optional
            _description_, by default "$p_{T}$ [GeV]"
        perf_var: str, optional
            The x axis variable, default is 'pt'
        h_line : float, optional
            draws a horizonatal line in the signal efficiency plot
        working_point: float, optional
            The working point to use for the plot. Only one out of
            [working_point, disc_cut, fixed_rejections] can be set
        disc_cut: float, optional
            The cut on the discriminant to use for the plot. Only one out of
            [working_point, disc_cut, fixed_rejections] can be set
        fixed_rejections: dict[Flavour, float]
            Show signal efficiency as a function of fixed background rejection. Only one
            out of [working_point, disc_cut, fixed_rejections] can be set
        **kwargs : kwargs
            key word arguments for `puma.VarVsEff`
        """
        # Check correct setting of working_point, disc_cut and fixed_rejections
        if sum([bool(working_point), bool(disc_cut), bool(fixed_rejections)]) > 1:
            raise ValueError("Only one of working_point, disc_cut, or fixed_rejections can be set")
        if not any([working_point, disc_cut, fixed_rejections]):
            raise ValueError("Either working_point or disc_cut must be set")

        # If fixed_rejections is given, call different function
        if fixed_rejections:
            self.plot_flat_rej_var_perf(
                fixed_rejections=fixed_rejections,
                suffix=suffix,
                perf_var=perf_var,
                h_line=h_line,
                **kwargs,
            )
            return

        # Split the kwargs according to if they are used for the plot or the curve
        var_perf_plot_kwargs = {
            "xlabel": xlabel,
            "n_ratio_panels": 1,
            "atlas_first_tag": self.atlas_first_tag,
            "atlas_second_tag": self.atlas_second_tag,
            "y_scale": 1.5,
            "logy": False,
        }

        # Update the default plot kwargs if present in kwargs and remove it from kwargs
        if kwargs is not None:
            # Init a list to loop over
            iter_kwargs_list = list(kwargs.keys())

            # Loop over the kwargs list
            for key in iter_kwargs_list:
                if key in var_perf_plot_kwargs:
                    var_perf_plot_kwargs[key] = kwargs.pop(key)

        # Init new var vs eff plot
        plot_sig_eff = VarVsEffPlot(
            mode="sig_eff", ylabel=self.signal.eff_str, **var_perf_plot_kwargs
        )

        # Adapt the atlas second tag
        plot_sig_eff.apply_modified_atlas_second_tag(
            self.signal,
            working_point=working_point,
            disc_cut=disc_cut,
            flat_per_bin=kwargs.get("flat_per_bin", False),
        )

        # Init new list for background plots
        plot_bkg = []

        # Loop over all backgrounds
        for background in self.backgrounds:
            # Init and append new background plot to the list
            plot_bkg.append(
                VarVsEffPlot(mode="bkg_rej", ylabel=background.rej_str, **var_perf_plot_kwargs)
            )

            # Adapt the atlas second label accordingly
            plot_bkg[-1].apply_modified_atlas_second_tag(
                self.signal,
                working_point=working_point,
                disc_cut=disc_cut,
                flat_per_bin=kwargs.get("flat_per_bin", False),
            )

        # Loop over the given taggers
        for tagger in self.taggers.values():
            # Load the discriminant values and check if
            discs = tagger.discriminant(self.signal)
            is_signal = tagger.is_flav(self.signal)

            # Assure that the variable is in the data for the given tagger
            assert perf_var in tagger.perf_vars, f"{perf_var} not in tagger {tagger.name} data!"

            # Add the variable to the plot
            plot_sig_eff.add(
                VarVsEff(
                    x_var_sig=tagger.perf_vars[perf_var][is_signal],
                    disc_sig=discs[is_signal],
                    label=tagger.label,
                    colour=tagger.colour,
                    working_point=working_point,
                    disc_cut=disc_cut,
                    **kwargs,
                ),
                reference=tagger.reference,
            )

            # Loop over the background plots and add the variables
            for counter, background in enumerate(self.backgrounds):
                is_bkg = tagger.is_flav(background)
                plot_bkg[counter].add(
                    VarVsEff(
                        x_var_sig=tagger.perf_vars[perf_var][is_signal],
                        disc_sig=discs[is_signal],
                        x_var_bkg=tagger.perf_vars[perf_var][is_bkg],
                        disc_bkg=discs[is_bkg],
                        label=tagger.label,
                        colour=tagger.colour,
                        working_point=working_point,
                        disc_cut=disc_cut,
                        **kwargs,
                    ),
                    reference=tagger.reference,
                )

        # Finalise the plot and draw it
        plot_sig_eff.draw()
        if h_line:
            plot_sig_eff.draw_hline(h_line)

        # Generate the name according to inputs and save the figure
        plot_base = "flat_per_bin" if kwargs.get("flat_per_bin") else "fixed_cut"
        wp_disc = (f"disc_cut{disc_cut}" if disc_cut else f"wp{int(working_point * 100)}").replace(
            ".", "p"
        )

        fname = f"{self.sig_str}eff_vs_{perf_var}_{plot_base}_{wp_disc}"
        suffix = f"{suffix}_" if suffix else ""
        self.save(plot_sig_eff, "profile", fname, suffix)

        # Save the background figures
        for counter, bkg in enumerate(self.backgrounds):
            plot_bkg[counter].draw()
            fname = f"{str(bkg)[0]}rej_vs_{perf_var}_{plot_base}_{wp_disc}"
            self.save(plot_bkg[counter], "profile", fname, suffix)

    def plot_flat_rej_var_perf(
        self,
        fixed_rejections: dict[Label, float],
        suffix: str | None = None,
        perf_var: str = "pt",
        h_line: float | None = None,
        **kwargs,
    ):
        """Plot signal efficiency as a function of a variable, with a fixed
        background rejection for each bin.

        Parameters
        ----------
        fixed_rejections : dict[Flavour, float]
            A dictionary of the fixed background rejections for each flavour, eg:
            fixed_rejections = {'cjets' : 0.1, 'ujets' : 0.01}
        suffix : str, optional
            suffix to add to output file name, by default None
        perf_var: str, optional
            The x axis variable, default is 'pt'
        h_line : float, optional
            draws a horizonatal line in the signal efficiency plot
        **kwargs : kwargs
            key word arguments for `puma.VarVsEff`
        """
        # Check for invalid inputs
        if inv_bkg := set(fixed_rejections.keys()) - {str(b) for b in self.backgrounds}:
            raise ValueError(f"Invalid background flavours: {inv_bkg}")
        if "disc_cut" in kwargs:
            raise ValueError("disc_cut should not be set for this plot")
        if "working_point" in kwargs:
            raise ValueError("working_point should not be set for this plot")

        # Split the kwargs according to if they are used for the plot or the curve
        var_perf_plot_kwargs = {
            "xlabel": r"$p_{T}$ [GeV]",
            "n_ratio_panels": 1,
            "atlas_first_tag": self.atlas_first_tag,
            "atlas_second_tag": self.atlas_second_tag,
            "y_scale": 1.5,
            "logy": False,
        }

        # Update the default plot kwargs if present in kwargs and remove it from kwargs
        if kwargs is not None:
            # Init a list to loop over
            iter_kwargs_list = list(kwargs.keys())

            # Loop over the kwargs list
            for key in iter_kwargs_list:
                if key in var_perf_plot_kwargs:
                    var_perf_plot_kwargs[key] = kwargs.pop(key)

        # Get a list of all backgrounds
        backgrounds = [Flavours[b] for b in fixed_rejections]

        # Init a list for background plots
        plot_bkg = []

        # Loop over all backgrounds
        for bkg in backgrounds:
            var_perf_plot_kwargs["atlas_second_tag"] = (
                f"{self.atlas_second_tag}\nConstant {bkg.rej_str.lower()} of"
                f" {fixed_rejections[bkg.name]} per bin"
            )

            # Init and append the background plot to the list
            plot_bkg.append(
                VarVsEffPlot(
                    mode="bkg_eff_sig_err",
                    ylabel=self.signal.eff_str,
                    **var_perf_plot_kwargs,
                )
            )

        # After all plots are created, loop over the taggers
        for tagger in self.taggers.values():
            # Get the disc values
            discs = tagger.discriminant(self.signal)
            is_signal = tagger.is_flav(self.signal)

            # Loop over the backgrounds
            for counter, bkg in enumerate(backgrounds):
                is_bkg = tagger.is_flav(bkg)

                # Check that variable is present in tagger data
                assert perf_var in tagger.perf_vars, f"{perf_var} not in tagger {tagger.name} data!"

                # We want x bins to all have the same background rejection, so we
                # select the plot mode as 'bkg_eff', and then treat the signal as
                # the background here. I.e, the API plots 'bkg_eff' on the y axis,
                # while keeping the 'sig_eff' a flat rate on the x axis, we therefore
                # pass the signal as the background, and the background as the
                # signal.
                plot_bkg[counter].add(
                    VarVsEff(
                        x_var_sig=tagger.perf_vars[perf_var][is_bkg],
                        disc_sig=discs[is_bkg],
                        x_var_bkg=tagger.perf_vars[perf_var][is_signal],
                        disc_bkg=discs[is_signal],
                        label=tagger.label,
                        colour=tagger.colour,
                        working_point=1 / fixed_rejections[bkg.name],
                        flat_per_bin=True,
                        **kwargs,
                    ),
                    reference=tagger.reference,
                )

        # Update the suffix
        suffix = f"_{suffix}" if suffix else ""

        # Loop over the backgrounds, draw and save all plots
        for counter, bkg in enumerate(backgrounds):
            plot_bkg[counter].draw()
            if h_line:
                plot_bkg[counter].draw_hline(h_line)
            details = f"{self.sig_str}eff_vs_{perf_var}_"
            base = f"{str(bkg)[0]}rej_flat_{int(fixed_rejections[bkg.name])}"
            self.save(plot_bkg[counter], "profile", details + base, suffix)

    def plot_fraction_scans(
        self,
        suffix: str | None = None,
        efficiency: float = 0.7,
        rej: bool = False,
        optimal_fc: bool = False,
        backgrounds: list[Label] | None = None,
        **kwargs,
    ):
        """Produce fraction scan (fc/fb) iso-efficiency plots.

        Parameters
        ----------
        suffix : str, optional
            suffix to add to output file name, by default None
        efficiency : float, optional
            signal efficiency, by default 0.7
        rej : bool, optional
            if True, plot rejection instead of efficiency, by default False
        optimal_fc : bool, optional
            if True, plot optimal fc/fb, by default False
        backgrounds : list[Flavour], optional
            List of background flavours, by default None, will use self.backgrounds
        **kwargs
            Keyword arguments for `puma.Line2DPlot
        """
        # Check signal flavour
        if self.signal not in {Flavours.bjets, Flavours.cjets}:
            raise ValueError("Signal flavour must be bjets or cjets")

        # Get the background flavours in a list
        backgrounds = (
            [Flavours[b] for b in backgrounds] if backgrounds is not None else self.backgrounds
        )

        # Check that only two background flavours are given
        if len(backgrounds) != 2:
            raise ValueError("Only two background flavours are supported")

        # Adapt the plot name and suffix accordingly
        frac = "fc" if self.signal == Flavours.bjets else "fb"
        back_str = "_".join([f.name for f in backgrounds])
        plot_name = f"{frac}_scan"
        suffix = combine_suffixes([f"{back_str}_eff{int(efficiency * 100)}", suffix])

        # Init a default kwargs dict
        fraction_scan_kwargs = {
            "logx": True,
            "logy": True,
        }

        # If kwargs are given, update the default dict accordingly
        if kwargs is not None:
            fraction_scan_kwargs.update(kwargs)

        # Define the granulartiy of the scan
        fxs = fraction_scan.get_fx_values(resolution=kwargs.pop("resolution", 100))

        # Adapt the tag
        tag = self.atlas_second_tag + "\n" if self.atlas_second_tag else ""
        tag += f"{self.signal.eff_str} = {efficiency:.0%}"
        tag += f"\n$f_{frac[1:]}$ scan"

        # Define a new plot for the scan
        plot = Line2DPlot(atlas_second_tag=tag, **kwargs)

        # Get good colours and define, if the efficiency or the rejection is calculated
        eff_or_rej = calc_eff if not rej else calc_rej
        colours = get_good_colours()

        # Loop over the taggers
        for counter, tagger in enumerate(self.taggers.values()):
            # Init zeros for both axes
            xs = np.zeros(len(fxs))
            ys = np.zeros(len(fxs))

            # Get the indices of the flavours
            sig_idx = tagger.is_flav(self.signal)
            bkg_1_idx = tagger.is_flav(backgrounds[0])
            bkg_2_idx = tagger.is_flav(backgrounds[1])

            # Loop over the fraction values
            for j, fx in enumerate(fxs):
                # Calculate disc values for the tagger for the given fraction values
                disc = tagger.discriminant(self.signal, fxs={frac: fx})

                # Calculate the effciency/rejection and add it to the value arrays
                xs[j] = eff_or_rej(disc[sig_idx], disc[bkg_1_idx], efficiency)
                ys[j] = eff_or_rej(disc[sig_idx], disc[bkg_2_idx], efficiency)

            # add curve for this tagger
            tagger_fx = tagger.fxs["fc" if self.signal == Flavours.bjets else "fb"]
            plot.add(
                Line2D(
                    x_values=xs,
                    y_values=ys,
                    label=f"{tagger.label} ($f_x={tagger_fx}$)",
                    colour=tagger.colour if tagger.colour else colours[counter],
                )
            )

            # Add a marker for the just added fraction scan
            # The is_marker bool tells the plot that this is a marker and not a line
            fx_idx = np.argmin(np.abs(fxs - tagger_fx))
            plot.add(
                Line2D(
                    x_values=xs[fx_idx],
                    y_values=ys[fx_idx],
                    marker="x",
                    markersize=15,
                    markeredgewidth=2,
                    colour=tagger.colour,
                ),
                is_marker=True,
            )

            # Plot optimal fc if wanted
            if optimal_fc:
                opt_idx, opt_fc = fraction_scan.get_optimal_fc(
                    np.stack((xs, ys), axis=1), fc_space=fxs, rej=rej
                )
                plot.add(
                    Line2D(
                        x_values=xs[opt_idx],
                        y_values=ys[opt_idx],
                        marker="x",
                        markersize=15,
                        markeredgewidth=1,
                        label=f"Optimal $f_x={opt_fc:.2f}$",
                    ),
                    is_marker=True,
                )

            # Adding labels
            if not rej:
                plot.xlabel = backgrounds[0].eff_str
                plot.ylabel = backgrounds[1].eff_str

            else:
                plot.xlabel = backgrounds[0].rej_str
                plot.ylabel = backgrounds[1].rej_str

        # Draw and save the plot
        plot.draw()
        self.save(plot, "scan", plot_name, suffix)

    def make_plot(self, plot_type: str, kwargs: dict):
        """Make a plot.

        Parameters
        ----------
        plot_type : str
            Type of plot
        kwargs : dict
            Keyword arguments for the plot
        """
        if plot_type not in self.plot_funcs:
            raise ValueError(f"Unknown plot type {plot_type}, choose from {self.plot_funcs.keys()}")
        self.plot_funcs[plot_type](**kwargs)



"""Support functions for general aux task related things."""

from __future__ import annotations


def get_aux_labels():
    """Get the truth labels for all aux tasks."""
    return {
        "vertexing": "ftagTruthVertexIndex",
        "track_origin": "ftagTruthOriginLabel",
    }


def get_trackOrigin_classNames():
    """Get the Track Origin class names.

    Returns
    -------
    `list` of `str`
        the class names
    """
    return [
        "Pileup",
        "Fake",
        "Primary",
        "FromB",
        "FromBC",
        "FromC",
        "FromTau",
        "OtherSecondary",
    ]



"""Unit test script for the functions in var_vs_eff.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
from ftag import Flavours
from matplotlib.testing.compare import compare_images

from puma import VarVsEff, VarVsEffPlot
from puma.utils.logging import logger, set_log_level

set_log_level(logger, "DEBUG")


class VarVsEffTestCase(unittest.TestCase):
    """Test class for the puma.var_vs_eff functions."""

    def setUp(self):
        self.working_point = 0.77
        self.disc_sig = np.linspace(-6, +6, 100)
        self.x_var_sig = np.exp(-self.disc_sig) * 10e3
        self.disc_bkg = np.linspace(-5.5, +6.6, 120)
        self.x_var_bkg = np.exp(-self.disc_bkg * 0.8) * 10e3 + 30

    def test_var_vs_eff_init_wrong_sig_shape(self):
        """Test var_vs_eff init."""
        with self.assertRaises(ValueError):
            VarVsEff(np.ones(4), np.ones(5))

    def test_var_vs_eff_init_wrong_bkg_shape(self):
        """Test var_vs_eff init."""
        with self.assertRaises(ValueError):
            VarVsEff(np.ones(6), np.ones(6), np.ones(4), np.ones(5))

    def test_var_vs_eff_init_flat_eff_disc_cut(self):
        """Test var_vs_eff init."""
        with self.assertRaises(ValueError):
            VarVsEff(
                np.ones(6),
                np.ones(6),
                flat_per_bin=True,
                disc_cut=1.0,
                working_point=0.77,
            )

    def test_var_vs_eff_init_flat_eff_no_wp(self):
        """Test var_vs_eff init."""
        with self.assertRaises(ValueError):
            VarVsEff(np.ones(6), np.ones(6), flat_per_bin=True, disc_cut=1.0)

    def test_var_vs_eff_init_disc_cut_wp(self):
        """Test var_vs_eff init."""
        with self.assertRaises(ValueError):
            VarVsEff(np.ones(6), np.ones(6), disc_cut=1.0, working_point=0.77)

    def test_var_vs_eff_init_no_disc_cut_no_wp(self):
        """Test var_vs_eff init."""
        with self.assertRaises(ValueError):
            VarVsEff(np.ones(6), np.ones(6))

    def test_var_vs_eff_init_disc_cut_wrong_shape(self):
        """Test var_vs_eff init."""
        with self.assertRaises(ValueError):
            VarVsEff(np.ones(6), np.ones(6), disc_cut=[1.0, 2.0])

    def test_var_vs_eff_set_bin_edges_list(self):
        """Test var_vs_eff _set_bin_edges."""
        var_plot = VarVsEff(
            x_var_sig=[0, 1, 2], disc_sig=[3, 4, 5], bins=[0, 1, 2], working_point=0.7
        )
        np.testing.assert_array_almost_equal(var_plot.bin_edges, [0, 1, 2])

    def test_var_vs_eff_set_bin_edges_only_signal(self):
        """Test var_vs_eff _set_bin_edges."""
        var_plot = VarVsEff(x_var_sig=[0, 1, 2], disc_sig=[3, 4, 5], bins=2, working_point=0.7)
        np.testing.assert_array_almost_equal(var_plot.bin_edges, [0, 1, 2], decimal=4)

    def test_var_vs_eff_set_bin_edges(self):
        """Test var_vs_eff _set_bin_edges."""
        var_plot = VarVsEff(
            x_var_sig=[0, 1, 2],
            disc_sig=[3, 4, 5],
            x_var_bkg=[-1, 1, 3],
            disc_bkg=[3, 4, 5],
            bins=2,
            working_point=0.7,
        )
        np.testing.assert_array_almost_equal(var_plot.bin_edges, [-1, 1, 3], decimal=4)

    def test_var_vs_eff_flat_eff_sig_eff(self):
        """Test var_vs_eff sig_eff."""
        n_bins = 4
        var_plot = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=n_bins,
        )
        np.testing.assert_array_almost_equal(
            var_plot.sig_eff[0], [self.working_point] * n_bins, decimal=2
        )

    def test_var_vs_eff_flat_eff_sig_rej(self):
        """Test var_vs_eff sig_rej."""
        n_bins = 4
        var_plot = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=n_bins,
        )
        np.testing.assert_array_almost_equal(
            var_plot.sig_rej[0], [1 / self.working_point] * n_bins, decimal=2
        )

    def test_var_vs_eff_one_bin(self):
        """Test var_vs_eff."""
        n_bins = 1
        var_plot = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=n_bins,
        )
        var_plot_comp = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            bins=n_bins,
        )
        var_plot_list = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            disc_cut=var_plot.disc_cut,
            bins=n_bins,
        )
        with self.subTest("Comparison var_plot_comp"):
            np.testing.assert_array_almost_equal(var_plot.sig_eff, var_plot_comp.sig_eff)
        with self.subTest("Comparison var_plot_list"):
            np.testing.assert_array_almost_equal(var_plot.sig_eff, var_plot_list.sig_eff)

    def test_var_vs_eff_divide_same(self):
        """Test var_vs_eff divide."""
        n_bins = 1
        var_plot = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=n_bins,
        )
        var_plot.y_var_mean, var_plot.y_var_std = var_plot.get("sig_eff")
        np.testing.assert_array_almost_equal(var_plot.divide(var_plot)[0], np.ones(1))

    def test_var_vs_eff_wrong_mode(self):
        """Test var_vs_eff wrong mode."""
        n_bins = 1
        var_plot = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=n_bins,
        )
        with self.assertRaises(ValueError):
            var_plot.y_var_mean, var_plot.y_var_std = var_plot.get("test")

    def test_var_vs_eff_divide_different_binning(self):
        """Test var_vs_eff divide."""
        var_plot = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=1,
        )
        var_plot_comp = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=2,
        )
        with self.assertRaises(ValueError):
            var_plot.y_var_mean, var_plot.y_var_std = var_plot.get("sig_eff")
            var_plot_comp.y_var_mean, var_plot_comp.y_var_std = var_plot.get("sig_eff")
            var_plot.divide(var_plot_comp)

    def test_var_vs_eff_eq_different_classes(self):
        """Test var_vs_eff eq."""
        var_plot = VarVsEff(x_var_sig=[0, 1, 2], disc_sig=[3, 4, 5], bins=2, working_point=0.7)
        self.assertNotEqual(var_plot, np.ones(6))

    def test_var_vs_eff_get(self):
        var_plot = VarVsEff(
            x_var_sig=self.disc_sig,
            disc_sig=self.x_var_sig,
            x_var_bkg=self.disc_bkg,
            disc_bkg=self.x_var_bkg,
            working_point=self.working_point,
            flat_per_bin=True,
            bins=1,
        )
        mode_options = ["sig_eff", "bkg_eff", "sig_rej", "bkg_rej"]
        for mode in mode_options:
            var_plot.get(mode)


class VarVsEffOutputTestCase(unittest.TestCase):
    """Test class for the puma.var_vs_eff_plot output."""

    def setUp(self):
        # Set up temp directory for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint:disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")
        # Generate discriminant and pT distribution for sig and bkg for two taggers
        # We want that both taggers yield the same discriminant values for the bkg
        # jets, a gaussian located at 0.
        # For low pT, we want both taggers to perform similar (discriminant values
        # are a gaussian located at 1).
        # For high pT (here pT>110), the "new" tagger signal is better separated from
        # bkg (discriminant values are then a gaussian located at 2)
        np.random.seed(42)
        n_random = 10_000

        # background (same for both taggers)
        self.disc_bkg = np.random.normal(loc=0, size=2 * n_random)
        self.x_var_bkg = np.random.uniform(0, 250, size=2 * n_random)

        # reference tagger (constant separation power for whole pT range)
        self.disc_sig_1 = np.random.normal(loc=1, size=2 * n_random)
        self.x_var_sig_1 = np.random.uniform(0, 250, size=2 * n_random)

        # new tagger (better separation for pT > 110)
        self.disc_sig_2 = np.concatenate((
            np.random.normal(loc=1, size=n_random),
            np.random.normal(loc=3, size=n_random),
        ))
        self.x_var_sig_2 = np.concatenate((
            np.random.uniform(0, 110, size=n_random),
            np.random.uniform(110, 250, size=n_random),
        ))

        # Define pT bins
        self.bins = [20, 30, 40, 60, 85, 110, 140, 175, 250]

    def test_var_vs_eff_plot_mode_option(self):
        with self.assertRaises(ValueError):
            VarVsEffPlot(
                mode="test",
                ylabel="Background rejection",
                xlabel=r"$p_{T}$ [GeV]",
                logy=True,
                atlas_second_tag="test",
                y_scale=1.5,
                n_ratio_panels=1,
                figsize=(9, 6),
            )

    def test_output_plot_flat_per_bin_bkg_rejection(self):
        """Test output plot with flat eff per bin - bkg rejection."""
        # define the curves
        ref_light = VarVsEff(
            x_var_sig=self.x_var_sig_1,
            disc_sig=self.disc_sig_1,
            x_var_bkg=self.x_var_bkg,
            disc_bkg=self.disc_bkg,
            bins=self.bins,
            working_point=0.5,
            disc_cut=None,
            flat_per_bin=True,
            label="reference model",
        )
        better_light = VarVsEff(
            x_var_sig=self.x_var_sig_2,
            disc_sig=self.disc_sig_2,
            x_var_bkg=self.x_var_bkg,
            disc_bkg=self.disc_bkg,
            bins=self.bins,
            working_point=0.5,
            disc_cut=None,
            flat_per_bin=True,
            linestyle="dashed",
            label="better model (by construction better for $p_T$ > 110)",
        )
        plot_bkg_rej = VarVsEffPlot(
            mode="bkg_rej",
            ylabel="Background rejection",
            xlabel=r"$p_{T}$ [GeV]",
            logy=True,
            atlas_second_tag=(
                "Unit test plot based on gaussian distributions. \n"
                "'reference model' should have bkg rejection of ~ 6.1 for the whole"
                " whole $p_T$ range\n"
                "'better model' should have bkg efficiency of ~ 6.1 for $p_T < 110$\n"
                "'better model' should have quite large bkg efficiency $p_T > 110$\n"
            ),
            y_scale=1.5,
            n_ratio_panels=1,
            figsize=(9, 6),
        )
        plot_bkg_rej.add(ref_light, reference=True)
        plot_bkg_rej.add(better_light)

        plot_bkg_rej.draw()

        name = "test_pt_dependence_rejection.png"
        plot_bkg_rej.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertEqual(
            None,
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            ),
        )

    def test_output_plot_flat_eff_bin_bkg_efficiency(self):
        """Test output plot with flat eff per bin."""
        # define the curves
        ref_light = VarVsEff(
            x_var_sig=self.x_var_sig_1,
            disc_sig=self.disc_sig_1,
            x_var_bkg=self.x_var_bkg,
            disc_bkg=self.disc_bkg,
            bins=self.bins,
            working_point=0.5,
            disc_cut=None,
            flat_per_bin=True,
            label="reference model",
        )
        better_light = VarVsEff(
            x_var_sig=self.x_var_sig_2,
            disc_sig=self.disc_sig_2,
            x_var_bkg=self.x_var_bkg,
            disc_bkg=self.disc_bkg,
            bins=self.bins,
            working_point=0.5,
            disc_cut=None,
            flat_per_bin=True,
            label="better model (by construction better for $p_T$ > 110)",
        )
        plot_bkg_rej = VarVsEffPlot(
            mode="bkg_eff",
            ylabel="background efficiency",
            xlabel=r"$p_{T}$ [GeV]",
            logy=False,
            atlas_second_tag=(
                "Unit test plot based on gaussian distributions. \n"
                "'reference model' should have bkg efficiency of ~ 0.16 "
                " whole $p_T$ range\n"
                "'better model' should have bkg efficiency of ~ 0.16 for $p_T < 110$\n"
                "'better model' should have bkg efficiency of ~ 0 for $p_T > 110$\n"
            ),
            y_scale=1.5,
            n_ratio_panels=1,
            figsize=(9, 6),
        )
        plot_bkg_rej.add(ref_light, reference=True)
        plot_bkg_rej.add(better_light)
        for cut in ref_light.disc_cut:
            print(cut)

        plot_bkg_rej.draw()

        name = "test_pt_dependence_bkg_efficiency.png"
        plot_bkg_rej.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertEqual(
            None,
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            ),
        )

    def test_var_vs_eff_info_str_fixed_eff(self):
        flat_wp_plot = VarVsEffPlot(
            mode="bkg_eff",
        )
        signal = Flavours["bjets"]

        flat_wp_plot.apply_modified_atlas_second_tag(signal=signal, working_point=0.7)

        expected_tag = "70.0% $b$-jet efficiency"
        self.assertEqual(flat_wp_plot.atlas_second_tag, expected_tag)

    def test_var_vs_eff_info_str_flat_wp(self):
        flat_wp_plot = VarVsEffPlot(
            mode="bkg_eff",
            atlas_second_tag="test",
        )
        signal = Flavours["bjets"]

        flat_wp_plot.apply_modified_atlas_second_tag(
            signal=signal, working_point=0.7, flat_per_bin=True
        )

        expected_tag = "test\nFlat 70.0% $b$-jet efficiency per bin"
        self.assertEqual(flat_wp_plot.atlas_second_tag, expected_tag)

    def test_var_vs_eff_info_str_fixed_disc(self):
        flat_wp_plot = VarVsEffPlot(
            mode="bkg_eff",
        )
        signal = Flavours["bjets"]

        flat_wp_plot.apply_modified_atlas_second_tag(signal=signal, disc_cut=3.0)

        expected_tag = "$D_{b}$ > 3.0"
        self.assertEqual(flat_wp_plot.atlas_second_tag, expected_tag)



# Hadron processing functions
from __future__ import annotations

import numpy as np


def GetOrderedHadrons(hadron_barcode, hadron_parent, n_max_showers=2):
    # This function orderes the hadron indices inside each jet in different showers
    # INPUTS:  f["truth_hadrons"]["barcode"], f["truth_hadrons"]["ftagTruthParentBarcode"]
    # n_showers max would be n_hadrons if there were 5 unrelated showers (not likely)

    # Output: Padded array of indices with shape (n_jets, n_showers, n_hadrons)

    n_jets, n_hadrons = hadron_barcode.shape

    set_parent_barcodes = [set(row[row > 0]) for row in hadron_parent]

    # Get indices for the parton shower within a family
    child_indices = [
        [
            [index for index, hadron in enumerate(hadron_list) if hadron == parent]
            for parent in parent_set
        ]
        for hadron_list, parent_set in zip(hadron_parent, set_parent_barcodes)
    ]
    parent_index = [
        [
            [index for index, hadron in enumerate(hadron_list) if hadron == parent]
            for parent in parent_set
        ]
        for hadron_list, parent_set in zip(hadron_barcode, set_parent_barcodes)
    ]
    family_indices = [
        [j + c[k] for k, j in enumerate(p)] for p, c in zip(parent_index, child_indices)
    ]

    # Now reshuffle them to keep always first the longest shower!
    #######################################################################################################
    # Calculate lengths for all family_indices
    lengths = np.array([len(family) for family in family_indices], dtype=object)

    # Create a mask for jets with more than one family member
    mask = lengths > 1

    # Create reshuffle indices only for those jets that have more than one family
    reshuffle = np.empty((n_jets, n_max_showers), dtype=object)  # n_max_shower

    reshuffle_indices = [
        np.argsort([len(elem) for elem in family_indices[i]])[::-1]
        for i in range(n_jets)
        if mask[i]
    ]
    reshuffle_indices = [idx[:n_max_showers] for idx in reshuffle_indices]
    if np.array(reshuffle_indices).size > 0:
        reshuffle[mask] = reshuffle_indices

    # Initialize sorted_family_indices
    sorted_family_indices = np.empty(n_jets, dtype=object)

    # Use the reshuffle to sort family_indices
    for i in range(n_jets):
        if mask[i]:  # Check if the jet has more than one family member
            sorted_indices = reshuffle[i]
            sorted_family_indices[i] = [family_indices[i][idx] for idx in sorted_indices]
        else:
            sorted_family_indices[i] = family_indices[i]  # Keep it unchanged if not applicable

    # Convert to numpy array if desired
    sorted_family_indices = np.array(sorted_family_indices, dtype=object)

    #######################################################################################################

    # Deal with the unrelated hadrons

    orphan_barcodes = np.where(
        (hadron_parent < 0) & (hadron_barcode > 0), hadron_barcode, 0
    )  # hadrons without parents
    unrelated_indices = [
        [
            [index]
            for index, hadron in enumerate(hadron_list)
            if (hadron not in parent_set) and (hadron > 0)
        ]
        for hadron_list, parent_set in zip(orphan_barcodes, set_parent_barcodes)
    ]  # remove parents
    extended_family_indices = [
        f + u for f, u in zip(sorted_family_indices.copy(), unrelated_indices)
    ]

    # Now select the more important (first) shower and pad the indices!

    # Initializethe padded array with -1
    padded_hadron_indices = np.full((n_jets, n_max_showers, n_hadrons), -1)

    # Fill the padded array
    for i, jet in enumerate(extended_family_indices):
        for j, shower in enumerate(jet):
            if j < n_max_showers:  # Limit the amount of showers
                hadrons_to_fill = shower[:n_hadrons]  # Limit to max number of hadrons
                padded_hadron_indices[i, j, : len(hadrons_to_fill)] = hadrons_to_fill

    return padded_hadron_indices


def AssociateTracksToHadron(track_parent, hadron_barcode, hadron_mask):
    # INPUTS #####
    # track_parent         shape (n_jets, n_tracks)
    # hadron_barcode       shape (n_jets, n_hadrons)
    # track_hadron_mask    shape (n_hadrons, n_jets, n_tracks)
    # OUTPUTS #####
    # track_to_hadron_array           shape (n_jets, n_hadrons, n_tracks)
    # inclusive_track_first_hadron    shape (n_jets, n_tracks)
    # inclusive_track_hadron          shape (n_jets, n_tracks)

    n_jets, n_tracks = track_parent.shape
    n_hadrons = hadron_barcode.shape[1]

    track_parent = np.where(
        track_parent < 0, np.nan, track_parent
    )  # use NAN so that they never match
    track_to_hadron_array = np.array([
        np.where(track_parent == hadron_barcode[:, k][:, np.newaxis], 1, 0)
        for k in range(n_hadrons)
    ])  # n_hadrons change to variable

    # build the inclusive vertex if needed
    inclusive_track_hadron = np.sum(track_to_hadron_array, axis=0)

    # Sum tracks from hadrons in the parton shower (applying the mask)
    track_hadron_mask = np.repeat(hadron_mask, n_tracks).reshape(n_hadrons, n_jets, n_tracks)
    inclusive_track_first_hadron = np.sum(
        np.where(track_hadron_mask, track_to_hadron_array, 0), axis=0
    )  # apply mask and sum across the first dimension

    # mask out hadrons with only one associated track
    mask_array = [
        np.repeat(
            np.where(np.sum(track_to_hadron_array[k], axis=1) >= 2, 1, 0)[:, np.newaxis], n_tracks
        ).reshape(n_jets, n_tracks)
        for k in range(n_hadrons)
    ]
    track_to_hadron_array = np.where(mask_array, track_to_hadron_array, mask_array)

    return np.array(track_to_hadron_array), inclusive_track_first_hadron, inclusive_track_hadron


def SelectHadron(truth_hadrons, hadron_index):
    invalid_jet_mask = hadron_index < 0

    # Select hadron with most tracks
    selected_hadron = truth_hadrons[np.arange(truth_hadrons.shape[0]), hadron_index.astype(int)]

    # Create a copy to preserve shape
    selected_hadron_copy = np.copy(selected_hadron)

    # Apply the mask and set invalid entries to np.nan (or you can set to 0)
    selected_hadron_copy[invalid_jet_mask] = -99  # Use np.nan or 0 based on preference

    return selected_hadron_copy


def select_tracks(track_hadron, index, element=0):
    rows = np.arange(track_hadron.shape[1])
    return track_hadron[index[rows, element], rows, :]



"""IntegratedEfficiency functions."""

from __future__ import annotations

import matplotlib as mpl
import numpy as np
from ftag import Flavours, Label

from puma.metrics import calc_eff
from puma.plot_base import PlotBase, PlotLineObject
from puma.utils import get_good_colours, get_good_linestyles, logger


class IntegratedEfficiency(PlotLineObject):
    """Represent a single IntegratedEfficiency curve."""

    def __init__(
        self,
        disc_sig: np.ndarray,
        disc_bkg: np.ndarray,
        key: str | None = None,
        n_vals: int = 500,
        tagger: str | None = None,
        flavour: str | Label = None,
        **kwargs,
    ) -> None:
        """Initialise properties of IntegratedEfficiency object.

        Parameters
        ----------
        disc_sig : np.array
            Discriminant values for signal
        disc_bkg : np.array
            Discriminant values for background
        key : str
            Identifier for IntegratedEfficiency e.g. tagger, by default None
        n_vals : int, optional
            Number of values to calculate the efficiency at, by default 500
        tagger : str, optional
            Tagger name, by default None
        flavour : str or Label, optional
            Flavour label of the jets, by default None
        **kwargs : kwargs
            Keyword arguments passed to `puma.PlotLineObject`

        Raises
        ------
        ValueError
            If `sig_eff` and `bkg_rej` have a different shape
        """
        super().__init__(**kwargs)
        self.disc_sig = np.asarray(disc_sig)
        self.disc_bkg = np.asarray(disc_bkg)
        self.n_vals = n_vals
        self.tagger = tagger
        self.key = key
        self.flavour = Flavours[flavour] if isinstance(flavour, str) else flavour
        if self.label is None and self.flavour is not None:
            self.label = self.flavour.label
        self._calc_profile()

    def _calc_profile(self):
        """Calculate the profile of the integrated efficiency curve."""
        self.eff, self.x = calc_eff(
            self.disc_sig,
            self.disc_bkg,
            np.linspace(0, 1, self.n_vals),
            return_cuts=True,
        )


class IntegratedEfficiencyPlot(PlotBase):
    """IntegratedEfficiencyPlot class."""

    def __init__(self, grid: bool = True, **kwargs) -> None:
        """IntegratedEfficiency plot properties.

        Parameters
        ----------
        grid : bool, optional
            Set the grid for the plots.
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`
        """
        super().__init__(grid=grid, **kwargs)
        self.int_effs = {}
        self.tagger_ls = {}
        self.label_colours = {}
        self.leg_tagger_labels = {}
        self.initialise_figure()
        self.disc_min, self.disc_max = (1e3, -1e3)
        self.default_linestyles = get_good_linestyles()
        self.legend_flavs = None
        self.leg_tagger_loc = "lower left"

        self.ymin = 0
        self.ymax = 1.2

    def add(self, int_eff: object, key: str | None = None):
        """Adding puma.Roc object to figure.

        Parameters
        ----------
        int_effs : puma.IntegratedEfficiency
            IntegratedEfficiency curve
        key : str, optional
            Unique identifier for IntegratedEfficiency curve, by default None

        Raises
        ------
        KeyError
            If unique identifier key is used twice
        """
        if key is None:
            key = len(self.int_effs) + 1
        if key in self.int_effs:
            raise KeyError(f"Duplicated key {key} already used for roc unique identifier.")

        self.int_effs[key] = int_eff
        # set linestyle
        if int_eff.tagger not in self.tagger_ls:
            self.tagger_ls[int_eff.tagger] = (
                self.default_linestyles[len(self.tagger_ls)]
                if int_eff.linestyle is None
                else int_eff.linestyle
            )
        elif int_eff.linestyle != self.tagger_ls[int_eff.tagger] and int_eff.linestyle is not None:
            logger.warning(
                "You specified a different linestyle for the same tagger"
                " %s. This will lead to a mismatch in the line colours"
                " and the legend.",
                int_eff.tagger,
            )
        if int_eff.linestyle is None:
            int_eff.linestyle = self.tagger_ls[int_eff.tagger]

        # set colours
        if int_eff.label not in self.label_colours:
            if int_eff.flavour is not None:
                self.label_colours[int_eff.label] = int_eff.flavour.colour
            else:
                curr_colours = set(self.label_colours.values())
                possible_colours = set(get_good_colours()) - curr_colours
                self.label_colours[int_eff.label] = (
                    possible_colours.pop() if int_eff.colour is None else int_eff.colour
                )
        elif int_eff.colour != self.label_colours[int_eff.label] and int_eff.colour is not None:
            logger.warning(
                "You specified a different colour for the same label"
                " %s. This will lead to a mismatch in the line colours"
                " and the legend.",
                int_eff.label,
            )
        if int_eff.colour is None:
            int_eff.colour = self.label_colours[int_eff.label]

    def get_xlim_auto(self):
        """Returns min and max efficiency values.

        Returns
        -------
        float
            Min and max efficiency values
        """
        for elem in self.int_effs.values():
            self.disc_min = min(np.min(elem.x), self.disc_min)
            self.disc_max = max(np.max(elem.x), self.disc_max)

        return self.disc_min, self.disc_max

    def make_legend(self, handles: list):
        """Make legend.

        Parameters
        ----------
        handles : list
            List of handles
        """
        line_list_tagger = [
            mpl.lines.Line2D(
                [],
                [],
                color="k",
                linestyle=self.tagger_ls[tagger],
                label=tagger,
            )
            for tagger in self.tagger_ls
        ]
        self.legend_flavs = self.axis_top.legend(
            handles=line_list_tagger,
            labels=[handle.get_label() for handle in line_list_tagger],
            loc=self.leg_tagger_loc,
            fontsize=self.leg_fontsize,
            ncol=self.leg_ncol,
        )
        self.axis_top.add_artist(self.legend_flavs)
        # Get the labels for the legends
        labels_list = []
        lines_list = []

        for line in handles:
            if line.get_label() not in labels_list:
                labels_list.append(line.get_label())
                lines_list.append(line)

        # Define the legend
        self.axis_top.legend(
            handles=lines_list,
            labels=labels_list,
            loc=self.leg_loc,
            fontsize=self.leg_fontsize,
            ncol=self.leg_ncol,
        )

    def draw(
        self,
        x_label: str = "Discriminant",
    ):
        """Draw plotting.

        Parameters
        ----------
        x_label : str, optional
            x-axis label, by default Discriminant
        """
        plt_handles = self.plot()
        xmin, xmax = self.get_xlim_auto()

        self.set_xlim(
            xmin if self.xmin is None else self.disc_min,
            xmax if self.xmax is None else self.disc_max,
        )
        self.set_title()
        self.set_y_lim()
        # self.set_log()
        self.set_y_lim()
        self.set_xlabel(label=x_label)
        self.set_ylabel(self.axis_top, label="Integrated efficiency")

        self.make_legend(plt_handles)

        self.plotting_done = True
        if self.apply_atlas_style is True:
            self.atlasify()
            # atlasify can only handle one legend. Therefore, we remove the frame of
            # the second legend by hand
            if self.legend_flavs is not None:
                self.legend_flavs.set_frame_on(False)

    def plot(self, **kwargs) -> mpl.lines.Line2D:
        """Plotting integrated efficiency curves.

        Parameters
        ----------
        **kwargs: kwargs
            Keyword arguments passed to plt.axis.plot

        Returns
        -------
        Line2D
            matplotlib Line2D object
        """
        plt_handles = []
        for key, elem in self.int_effs.items():
            plt_handles += self.axis_top.plot(
                elem.x,
                elem.eff,
                linestyle=elem.linestyle,
                color=elem.colour,
                label=elem.label if elem is not None else key,
                zorder=2,
                **kwargs,
            )
        return plt_handles



"""Dummy data generation for plotting."""

from __future__ import annotations

from tempfile import NamedTemporaryFile

import h5py
import numpy as np
import pandas as pd
from ftag.mock import softmax
from numpy.lib.recfunctions import structured_to_unstructured as s2u
from numpy.lib.recfunctions import unstructured_to_structured as u2s


def get_dummy_multiclass_scores(
    size: int = 9_999,
    bjets_mean: float = 1.4,
    seed: int = 42,
):
    """
    Generate dummy data representing output of 3 class classifier.
    Adapted to light-, c- and b-jets and values of `HadronConeExclTruthLabelID`.

    Parameters
    ----------
    size : int, optional
        Size of dummy data. For each of the 3 classes, the same amount is produced,
        if size cannot be divided by 3, next smaller number is taken,
        by default 9_999
    bjets_mean : float, optional
        Mean value of the b-jets 3D gaussian, the more away from 0, the better the
        b-tagging performance, by default 1.4
    seed : int, optional
        Random seed for number generation, by default 42

    Returns
    -------
    np.ndarray
        Output scores of the shape (size, 3)
    np.ndarray
        Labels of shape (size,). The order of the output is light-jets, c-jets, b-jets

    """
    size_class = int(size / 3)
    rng = np.random.default_rng(seed=seed)
    ujets = softmax(rng.normal(loc=[-1, 0, 0], scale=1, size=(size_class, 3)), axis=1)
    cjets = softmax(rng.normal(loc=[0, 1, 0], scale=2.5, size=(size_class, 3)), axis=1)
    bjets = softmax(rng.normal(loc=[0, 0, bjets_mean], scale=2, size=(size_class, 3)), axis=1)
    output = np.concatenate((ujets, cjets, bjets))
    output = u2s(output, dtype=np.dtype([("ujets", "f4"), ("cjets", "f4"), ("bjets", "f4")]))
    labels = np.concatenate((
        np.zeros(size_class),
        np.ones(size_class) * 4,
        np.ones(size_class) * 5,
    ))
    return output, labels


def get_dummy_2_taggers(
    size: int = 9_999,
    shuffle: bool = True,
    seed: int = 42,
    add_pt: bool = False,
    label: str = "HadronConeExclTruthLabelID",
    return_file: bool = False,
):
    """
    Wrapper function of `get_dummy_multiclass_scores` to generate classifier output
    for 2 taggers, in this case rnnip and dips as well as HadronConeExclTruthLabelID.


    Parameters
    ----------
    size : int, optional
        Size of dummy data, by default 9_999
    shuffle : bool, optional
        If True shuffles the dummy data, by default True
    seed : int, optional
        Random seed for number generation (will count +10 for second tagger),
        by default 42
    add_pt : bool, optional
        Specify if pt column should be added as well, by default False
    label : str, optional
        Name of the label column, by default "HadronConeExclTruthLabelID"
    return_file : bool, optional
        If True returns a file object instead of a pandas.DataFrame,

    Returns
    -------
    df_gen : pandas.DataFrame or file
        Dataframe with columns
        [label, rnnip_pu, rnnip_pc, rnnip_pb, dips_pu, dips_pc,
        dips_pb] if `add_pt` is True also pt is added
    """
    output_rnnip, labels = get_dummy_multiclass_scores(bjets_mean=0.9, size=size, seed=seed)
    df_gen = pd.DataFrame(s2u(output_rnnip), columns=["rnnip_pu", "rnnip_pc", "rnnip_pb"])
    df_gen[label] = labels
    output_dips, _ = get_dummy_multiclass_scores(bjets_mean=1.4, size=size, seed=seed + 10)
    df_gen2 = pd.DataFrame(s2u(output_dips), columns=["dips_pu", "dips_pc", "dips_pb"])
    df_gen = pd.concat([df_gen, df_gen2], axis=1)
    if add_pt:
        rng = np.random.default_rng(seed=seed)
        df_gen["pt"] = rng.exponential(100_000, size=len(df_gen))
    if shuffle:
        df_gen = df_gen.sample(frac=1).reset_index(drop=True)

    df_gen["n_truth_promptLepton"] = 0

    if return_file:
        fname = NamedTemporaryFile(  # pylint: disable=R1732
            mode="w", suffix=".h5", delete=False
        ).name
        file = h5py.File(fname, "w")
        file.create_dataset(name="jets", data=df_gen.to_records())
        return file

    return df_gen.to_records()


def get_dummy_tagger_aux(
    size: int = 9_999,
    n_tracks: int = 40,
    shuffle: bool = True,
    seed: int = 42,
    label: str = "HadronConeExclTruthLabelID",
):
    """
    Function to generate aux task output for a tagger, in this case GN2 with
    vertexing and track origin classification. Also includes jet level
    classifier output and HadronConeExclTruthLabelID.


    Parameters
    ----------
    size : int, optional
        Size of dummy data, by default 9_999
    n_tracks : int, optional
        Number of tracks per jet in dummy data, by default 40
    shuffle : bool, optional
        If True shuffles the dummy data, by default True
    seed : int, optional
        Random seed for number generation (will count +10 for second tagger),
        by default 42
    label : str, optional
        Name of the label column, by default "HadronConeExclTruthLabelID"
    vtx_label_var : str, optional
        Name of the truth vertex label, by default "ftagTruthVertexIndex"

    Returns
    -------
    df_gen : file
        h5 file with "jets" and "tracks" datasets
    """
    output_gn2, labels = get_dummy_multiclass_scores(bjets_mean=0.9, size=size, seed=seed)
    df_gen = pd.DataFrame(s2u(output_gn2), columns=["GN2_pu", "GN2_pc", "GN2_pb"])
    df_gen[label] = labels

    rng = np.random.default_rng(seed=seed)
    df_gen["pt"] = rng.exponential(100_000, size=len(df_gen))
    df_gen["eta"] = rng.normal(0, 2, size=len(df_gen))
    if shuffle:
        df_gen = df_gen.sample(frac=1).reset_index(drop=True)

    df_gen["n_truth_promptLepton"] = 0

    track_pt = rng.exponential(1000, size=(len(df_gen), n_tracks))
    track_eta = rng.normal(0, 2, size=(len(df_gen), n_tracks))
    track_deta = rng.normal(0, 0.5, size=(len(df_gen), n_tracks))
    track_dphi = rng.uniform(-np.pi, np.pi, size=(len(df_gen), n_tracks))
    vtx_labels = np.fabs(
        np.rint(rng.normal(loc=0, scale=10, size=(len(df_gen), n_tracks))).astype(int)
    )
    vtx_reco = np.fabs(
        np.rint(rng.normal(loc=0, scale=10, size=(len(df_gen), n_tracks))).astype(int)
    )
    trk_or_labels = rng.choice(8, size=(len(df_gen), n_tracks)).astype(int)
    trk_or_reco = rng.choice(8, size=(len(df_gen), n_tracks)).astype(int)
    aux_dtype = np.dtype([
        ("pt", "f4"),
        ("eta", "f4"),
        ("deta", "f4"),
        ("dphi", "f4"),
        ("ftagTruthVertexIndex", "i4"),
        ("GN2_aux_VertexIndex", "i4"),
        ("ftagTruthOriginLabel", "i4"),
        ("GN2_aux_TrackOrigin", "i4"),
    ])
    aux_info = np.rec.fromarrays(
        [
            track_pt,
            track_eta,
            track_deta,
            track_dphi,
            vtx_labels,
            vtx_reco,
            trk_or_labels,
            trk_or_reco,
        ],
        dtype=aux_dtype,
    )

    fname = NamedTemporaryFile(  # pylint: disable=R1732
        mode="w", suffix=".h5", delete=False
    ).name
    file = h5py.File(fname, "w")
    file.create_dataset(name="jets", data=df_gen.to_records())
    file.create_dataset(name="tracks", data=aux_info)

    return fname, file



"""Unit test script for the functions in plot_base.py."""

from __future__ import annotations

import unittest

from puma import PlotBase, PlotObject
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class PlotObjectTestCase(unittest.TestCase):
    """Test class for the puma.PlotObject dataclass."""

    def test_only_one_input_figsize(self):
        """Test only one input figsize."""
        with self.assertRaises(ValueError):
            PlotObject(figsize=1)

    def test_only_tuple_three_inputs_figsize(self):
        """Test only tuple three inputs figsize."""
        with self.assertRaises(ValueError):
            PlotObject(figsize=(1, 2, 3))

    def test_tuple_input_figsize(self):
        """Test tuple input figsize."""
        figsize = PlotObject(figsize=(1, 2)).figsize
        self.assertEqual(figsize, (1, 2))

    def test_list_input_figsize(self):
        """Test list input figsize."""
        figsize = PlotObject(figsize=[1, 2]).figsize
        self.assertEqual(figsize, (1, 2))

    def test_list_input_wrong_len_figsize(self):
        """Test list input wrong len figsize."""
        with self.assertRaises(ValueError):
            PlotObject(figsize=[1, 2, 3])

    def test_wrong_n_ratio_panels(self):
        """Test wrong n ratio panels."""
        with self.assertRaises(ValueError):
            PlotObject(n_ratio_panels=5)

    def test_wrong_ymin_ratio(self):
        """Test wrong ymin_ratio."""
        with self.assertRaises(ValueError):
            PlotObject(n_ratio_panels=2, ymin_ratio=[0])

    def test_wrong_ymax_ratio(self):
        """Test wrong ymax_ratio."""
        with self.assertRaises(ValueError):
            PlotObject(n_ratio_panels=2, ymax_ratio=[0])

    def test_set_ratio_label_invalid_ratio_panel(self):
        plot_object = PlotBase(n_ratio_panels=2)
        plot_object.set_ratio_label(ratio_panel=1, label="Label")
        plot_object.set_ratio_label(ratio_panel=2, label="Label")
        with self.assertRaises(ValueError):
            plot_object.set_ratio_label(ratio_panel=3, label="Label")


class PlotBaseTestCase(unittest.TestCase):
    """Test class for the puma.PlotBase class."""

    def test_ymin_ymax_ratio(self):
        """Test correct ymax/ymin_ratio."""
        plot_object = PlotBase(n_ratio_panels=2, ymin_ratio=[0, 1], ymax_ratio=[1, 2])
        plot_object.initialise_figure()
        plot_object.set_y_lim()
        for i in range(2):
            ymin, ymax = plot_object.ratio_axes[i].get_ylim()
            self.assertEqual((ymin, ymax), (i, i + 1))



"""Unit test script for the functions in histogram.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

from matplotlib.testing.compare import compare_images

from puma import PiePlot
from puma.utils import get_good_colours, logger, set_log_level

set_log_level(logger, "DEBUG")


class PiePlotTestCase(unittest.TestCase):
    """Test class for puma.PiePlot."""

    def setUp(self):
        # Set up directories for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")

    def test_plot_pie_chart_default_style(self):
        """Check if pie chart is plotted correctly (using default style)."""
        pie_plot = PiePlot(
            wedge_sizes=[20, 40, 30, 10],
            labels=["light-flavour jets", "c-jets", "b-jets", "tau-jets"],
        )
        name = "test_pie_chart_default_style.png"
        pie_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_plot_pie_chart_custom_style(self):
        """Check if pie chart is plotted correctly (using default style)."""
        pie_plot = PiePlot(
            wedge_sizes=[20, 40, 30, 10],
            labels=["light-flavour jets", "c-jets", "b-jets", "tau-jets"],
            draw_legend=True,
            colours=get_good_colours()[:4],
            # have a look at the possible kwargs for matplotlib.pyplot.pie here:
            # https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.pie.html
            mpl_pie_kwargs={
                "explode": [0, 0, 0, 0.1],
                "shadow": False,
                "startangle": 90,
                "textprops": {"fontsize": 10},
                "radius": 1,
                "wedgeprops": {"width": 0.4, "edgecolor": "w"},
                "pctdistance": 0.4,
            },
            # kwargs passed to puma.PlotObject
            atlas_second_tag=("Unit test plot to test if the custom\nstyling of the pie plot"),
            figsize=(5.5, 3.5),
            y_scale=1.3,
        )
        name = "test_pie_chart_custom_style.png"
        pie_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # pie_plot.savefig(f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )



"""Unit test script for the functions in roc.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
import py
from matplotlib.testing.compare import compare_images

from puma import Roc, RocPlot
from puma.utils.logging import logger, set_log_level

set_log_level(logger, "DEBUG")


class RocTestCase(unittest.TestCase):
    """Test class for the puma.roc functions."""

    def setUp(self):
        self.sig_eff = np.linspace(0.4, 1, 100)
        self.bkg_rej = np.exp(-self.sig_eff) * 10e3

    def test_roc_init(self):
        """Test roc init."""
        with self.assertRaises(ValueError):
            Roc(np.ones(4), np.ones(5))

    def test_ratio_same_object(self):
        """Test roc divide function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej)
        roc_curve_ref = Roc(self.sig_eff, self.bkg_rej)
        _, ratio, _ = roc_curve.divide(roc_curve_ref)

        np.testing.assert_array_almost_equal(ratio, np.ones(len(self.bkg_rej)))

    def test_ratio_factor_two(self):
        """Test roc divide function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej)
        roc_curve_ref = Roc(self.sig_eff, self.bkg_rej * 2)
        _, ratio, _ = roc_curve.divide(roc_curve_ref)

        np.testing.assert_array_almost_equal(ratio, 1 / 2 * np.ones(len(self.bkg_rej)))

    def test_ratio_different_sig_interval(self):
        """Test roc divide function."""
        sig_eff = np.linspace(0.4, 0.9, 6)
        sig_eff_ref = np.linspace(0.6, 1, 5)
        bkg_rej = np.exp(-sig_eff) * 10e3
        bkg_rej_ref = np.exp(-sig_eff_ref) * 10e3
        roc_curve = Roc(sig_eff, bkg_rej)
        roc_curve_ref = Roc(sig_eff_ref, bkg_rej_ref * 2)
        with py.test.raises(ValueError, match="Signal efficiencies"):
            _, _, _ = roc_curve.divide(roc_curve_ref)

    def test_ratio_factor_two_inverse(self):
        """Test roc divide function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej)
        roc_curve_ref = Roc(self.sig_eff, self.bkg_rej * 2)
        _, ratio, _ = roc_curve.divide(roc_curve_ref, inverse=True)

        np.testing.assert_array_almost_equal(ratio, 2 * np.ones(len(self.bkg_rej)))

    def test_binomial_error_no_ntest(self):
        """Test roc binomial_error function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej)
        with self.assertRaises(ValueError):
            roc_curve.binomial_error()

    def test_binomial_error_only_zeros(self):
        """Test roc binomial_error function."""
        roc_curve = Roc(self.sig_eff, np.zeros(len(self.sig_eff)), n_test=10e5)
        np.testing.assert_array_almost_equal(roc_curve.binomial_error(), [])

    def test_binomial_error_example(self):
        """Test roc binomial_error function."""
        error_rej = np.array([8.717798, 35.0, 99.498744])
        roc_curve = Roc(np.array([0.1, 0.2, 0.3]), np.array([20, 50, 100]), n_test=100)
        np.testing.assert_array_almost_equal(roc_curve.binomial_error(), error_rej)

    def test_binomial_error_example_norm(self):
        """Test roc binomial_error function."""
        error_rej = np.array([8.717798, 35.0, 99.498744]) / np.array([20, 50, 100])
        roc_curve = Roc(np.array([0.1, 0.2, 0.3]), np.array([20, 50, 100]), n_test=100)
        np.testing.assert_array_almost_equal(roc_curve.binomial_error(norm=True), error_rej)

    def test_binomial_error_example_pass_ntest(self):
        """Test roc binomial_error function."""
        error_rej = np.array([8.717798, 35.0, 99.498744])
        roc_curve = Roc(np.array([0.1, 0.2, 0.3]), np.array([20, 50, 100]))
        np.testing.assert_array_almost_equal(roc_curve.binomial_error(n_test=100), error_rej)


class RocMaskTestCase(unittest.TestCase):
    """Test class for the puma.roc non_zero_mask function."""

    def setUp(self):
        self.sig_eff = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])
        self.bkg_rej = np.array([0, 0.2, 0, 0.4, 0.5, 0, 0.7])

    def test_non_zero_mask(self):
        """Test roc non_zero_mask function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej)
        np.testing.assert_array_almost_equal(
            roc_curve.non_zero_mask, [False, True, False, True, True, False, True]
        )

    def test_non_zero_mask_xmin(self):
        """Test roc non_zero_mask function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej, xmin=0.4)
        np.testing.assert_array_almost_equal(
            roc_curve.non_zero_mask, [False, False, False, True, True, False, True]
        )

    def test_non_zero_mask_xmax(self):
        """Test roc non_zero_mask function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej, xmax=0.6)
        np.testing.assert_array_almost_equal(
            roc_curve.non_zero_mask, [False, True, False, True, True, False, False]
        )

    def test_non_zero_mask_xmin_xmax(self):
        """Test roc non_zero_mask function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej, xmax=0.6, xmin=0.4)
        np.testing.assert_array_almost_equal(
            roc_curve.non_zero_mask, [False, False, False, True, True, False, False]
        )

    def test_non_zero(self):
        """Test roc non_zero function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej)
        result_bkg_rej = self.bkg_rej[[False, True, False, True, True, False, True]]
        result_sig_eff = self.sig_eff[[False, True, False, True, True, False, True]]
        np.testing.assert_array_almost_equal(roc_curve.non_zero, (result_bkg_rej, result_sig_eff))

    def test_non_zero_xmin(self):
        """Test roc non_zero function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej, xmin=0.4)
        result_bkg_rej = self.bkg_rej[[False, False, False, True, True, False, True]]
        result_sig_eff = self.sig_eff[[False, False, False, True, True, False, True]]
        np.testing.assert_array_almost_equal(roc_curve.non_zero, (result_bkg_rej, result_sig_eff))

    def test_non_zero_xmax(self):
        """Test roc non_zero function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej, xmax=0.6)
        result_bkg_rej = self.bkg_rej[[False, True, False, True, True, False, False]]
        result_sig_eff = self.sig_eff[[False, True, False, True, True, False, False]]
        np.testing.assert_array_almost_equal(roc_curve.non_zero, (result_bkg_rej, result_sig_eff))

    def test_non_zero_xmin_xmax(self):
        """Test roc non_zero function."""
        roc_curve = Roc(self.sig_eff, self.bkg_rej, xmax=0.6, xmin=0.4)
        result_bkg_rej = self.bkg_rej[[False, False, False, True, True, False, False]]
        result_sig_eff = self.sig_eff[[False, False, False, True, True, False, False]]
        np.testing.assert_array_almost_equal(roc_curve.non_zero, (result_bkg_rej, result_sig_eff))


class RocOutputTestCase(unittest.TestCase):
    """Test class for the puma.roc_plot function."""

    def setUp(self):
        # Set up temp directory for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")

        self.sig_eff = np.array([0.5, 0.56, 0.63, 0.76, 0.83, 0.84, 0.85, 0.88, 0.9, 0.93, 1])
        self.u_rej_1 = np.array([1542, 918, 426, 78, 31, 26, 22, 12, 8, 4, 1])
        self.u_rej_2 = np.array([3061, 1642, 709, 112, 41, 34, 28, 15, 9, 4, 1])
        self.c_rej_1 = np.array([26, 17, 10, 4.1, 2.9, 2.7, 2.6, 2.2, 1.9, 1.6, 1])
        self.c_rej_2 = np.array([45, 27, 14, 4, 3, 2.9, 2.7, 2.2, 2.0, 1.63, 1])

        # sig efficiency with different start
        self.sig_eff_short = np.linspace(0.5, 1, 100)

    def test_output_two_curves_no_ratio(self):
        """Test with two curves of same flavour, without ratio panel."""
        plot = RocPlot(
            n_ratio_panels=0,
            ylabel="Light-jet rejection",
            xlabel="$b$-jet efficiency",
            atlas_second_tag=(
                "$\\sqrt{s}=13$ TeV, PFlow Jets\n$t\\bar{t}$ dummy sample," " $f_{c}=0.018$"
            ),
            y_scale=1.5,
        )

        # Add two roc curves
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_1,
                rej_class="ujets",
                label="reference",
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_2,
                rej_class="ujets",
                label="test",
            )
        )

        # Draw the figure
        plot.draw()

        name = "test_roc_two_curves_no_ratio.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copyfile(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=2.5,
            )
        )

    def test_output_two_curves_one_ratio(self):
        """Test with two curves of same flavour, one ratio panel."""
        plot = RocPlot(
            n_ratio_panels=1,
            ylabel="Light-jet rejection",
            xlabel="$b$-jet efficiency",
            atlas_second_tag=(
                "$\\sqrt{s}=13$ TeV, PFlow Jets\n$t\\bar{t}$ dummy sample," " $f_{c}=0.018$"
            ),
            y_scale=1.5,
            # logy=False,
        )

        # Add two roc curves
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_1,
                rej_class="ujets",
                label="reference",
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_2,
                rej_class="ujets",
                label="test",
            )
        )

        plot.set_ratio_class(1, "ujets")

        # Draw the figure
        plot.draw()

        name = "test_roc_two_curves_1_ratio.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copyfile(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=2.5,
            )
        )

    def test_output_two_curves_one_ratio_uncertainties(self):
        """Test with two curves of same flavour, one ratio panel."""
        plot = RocPlot(
            n_ratio_panels=1,
            ylabel="Light-jet rejection",
            xlabel="$b$-jet efficiency",
            atlas_second_tag=(
                "$\\sqrt{s}=13$ TeV, PFlow Jets\n$t\\bar{t}$ dummy sample," " $f_{c}=0.018$"
            ),
            y_scale=1.5,
            # logy=False,
        )

        # Add two roc curves
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_1 * 2,
                rej_class="ujets",
                label="reference",
                n_test=1_000,
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_2 * 2,
                rej_class="ujets",
                label="test",
                n_test=1_000,
            )
        )

        plot.set_ratio_class(1, "ujets")

        # Draw the figure
        plot.draw()

        name = "test_roc_two_curves_1_ratio_unc.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copyfile(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=2.5,
            )
        )

    def test_output_four_curves_two_ratio(self):
        """Test with two curves for each flavour, two ratio panels."""
        plot = RocPlot(
            n_ratio_panels=2,
            ylabel="Background rejection",
            xlabel="$b$-jet efficiency",
            atlas_second_tag=(
                "$\\sqrt{s}=13$ TeV, PFlow Jets\n$t\\bar{t}$ dummy sample," " $f_{c}=0.018$"
            ),
        )

        # Add four roc curves
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_1,
                rej_class="ujets",
                label="reference",
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_2,
                rej_class="ujets",
                label="test",
            )
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.c_rej_1,
                rej_class="cjets",
                label="reference",
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.c_rej_2,
                rej_class="cjets",
                label="test",
            )
        )

        plot.set_ratio_class(1, "ujets")
        plot.set_ratio_class(2, "cjets")

        # Draw the figure
        plot.draw()

        name = "test_roc_four_curves_2_ratio.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copyfile(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=2.5,
            )
        )

    def test_output_ratio_legend_four_curves_two_ratio(self):
        """Test with two curves for each flavour, two ratio panels, and ratio legend."""
        plot = RocPlot(
            n_ratio_panels=2,
            ylabel="Background rejection",
            xlabel="$b$-jet efficiency",
            atlas_second_tag=(
                "$\\sqrt{s}=13$ TeV, PFlow Jets\n$t\\bar{t}$ dummy sample," " $f_{c}=0.018$"
            ),
        )

        # Add four roc curves
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_1,
                rej_class="ujets",
                label="reference",
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_2,
                rej_class="ujets",
                label="test",
            )
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.c_rej_1,
                rej_class="cjets",
                label="reference",
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.c_rej_2,
                rej_class="cjets",
                label="test",
            )
        )

        plot.set_ratio_class(1, "ujets")
        plot.set_ratio_class(2, "cjets")

        # Draw the figure
        plot.draw()

        name = "test_output_ratio_legend_four_curves_two_ratio.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copyfile(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=2.5,
            )
        )

    def test_output_four_curves_two_ratio_uncertainties(self):
        """Test with two curves for each flavour, two ratio panels and binom. unc."""
        plot = RocPlot(
            n_ratio_panels=2,
            ylabel="Background rejection",
            xlabel="$b$-jet efficiency",
            atlas_second_tag=(
                "$\\sqrt{s}=13$ TeV, PFlow Jets\n$t\\bar{t}$ dummy sample," " $f_{c}=0.018$"
            ),
        )

        # Add four roc curves
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_1,
                rej_class="ujets",
                label="reference",
                n_test=1_000,
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.u_rej_2,
                rej_class="ujets",
                label="test",
                n_test=1_000,
            )
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.c_rej_1,
                rej_class="cjets",
                label="reference",
                n_test=1_000,
            ),
            reference=True,
        )
        plot.add_roc(
            Roc(
                self.sig_eff,
                self.c_rej_2,
                rej_class="cjets",
                label="test",
                n_test=1_000,
            )
        )

        plot.set_ratio_class(1, "ujets")
        plot.set_ratio_class(2, "cjets")

        # Draw the figure
        plot.draw()

        name = "test_roc_four_curves_2_ratio_unc.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copyfile(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=2.5,
            )
        )



"""Unit test script for the functions in utils/vertexing.py."""

from __future__ import annotations

import unittest

import numpy as np

from puma.utils import logger, set_log_level
from puma.utils.vertexing import (
    associate_vertices,
    build_vertices,
    calculate_vertex_metrics,
    clean_indices,
    clean_reco_vertices,
    clean_truth_vertices,
)

set_log_level(logger, "DEBUG")


class CleanIndicesTestCase(unittest.TestCase):
    """Test case for clean_indices function."""

    def test_remove(self):
        """Check case where indices are removed."""
        vertex_ids = np.array([[0, 1, 1, 2, 1]])
        condition = np.array([[True, False, False, False, True]])
        updated_ids = clean_indices(vertex_ids, condition, mode="remove")
        expected_result = np.array([[-99, 1, 1, 2, -99]])
        np.testing.assert_array_equal(updated_ids, expected_result)

    def test_merge(self):
        """Check case where indices are merged."""
        vertex_ids = np.array([[0, 1, 1, 2, 1]])
        condition = np.array([[True, False, False, True, False]])
        updated_ids = clean_indices(vertex_ids, condition, mode="merge")
        expected_result = np.array([[3, 1, 1, 3, 1]])
        np.testing.assert_array_equal(updated_ids, expected_result)

    def test_invalid_mode(self):
        """Check case where an invalid mode is provided."""
        vertex_ids = np.array([[0, 1, 1, 2, 1]])
        condition = np.array([[True, False, False, False, True]])
        with self.assertRaises(ValueError):
            clean_indices(vertex_ids, condition, mode="invalid")


class BuildVerticesTestCase(unittest.TestCase):
    """Test case for build_vertices function."""

    def test_no_vertices(self):
        """Check case where there are no vertices."""
        indices = np.array([0, 1, 2, 3, 4])
        vertices = build_vertices(indices)
        self.assertEqual(vertices.shape, (0, indices.shape[0]))

    def test_simple_case(self):
        """Check simple case with one vertex."""
        indices = np.array([0, 1, 2, 2, 3])
        expected_result = np.array([[False, False, True, True, False]])
        vertices = build_vertices(indices)
        np.testing.assert_array_equal(vertices, expected_result)

    def test_complex_case(self):
        """Check complex case with multiple vertices."""
        indices = np.array([0, 1, 2, 2, 3, 1, 3, 3, 4, 1])
        expected_result = np.array([
            [False, True, False, False, False, True, False, False, False, True],
            [False, False, True, True, False, False, False, False, False, False],
            [False, False, False, False, True, False, True, True, False, False],
        ])
        vertices = build_vertices(indices)
        np.testing.assert_array_equal(vertices, expected_result)

    def test_arbitrary_indices(self):
        """Check case where vertex indices are arbitrary."""
        indices = np.array([22, 19, 22, 103, 103])
        expected_result = np.array([
            [True, False, True, False, False],
            [False, False, False, True, True],
        ])
        vertices = build_vertices(indices)
        np.testing.assert_array_equal(vertices, expected_result)

    def test_ignore_negatives(self):
        """Check case where negative indices are present."""
        indices = np.array([0, -2, 0, -2, -1, -1, -1])
        expected_result = np.array([[True, False, True, False, False, False, False]])
        vertices = build_vertices(indices)
        np.testing.assert_array_equal(vertices, expected_result)


class AssociateVerticesTestCase(unittest.TestCase):
    """Test case for associate_vertices function."""

    def test_no_associations(self):
        """Check case where there are no associations to make."""
        vertices1 = np.array([
            [True, True, False, False, False],
        ])
        vertices2 = np.array([
            [False, False, True, True, False],
        ])
        expected_assoc = np.array([[False]])
        expected_common = np.array([[0]])
        assoc, common = associate_vertices(vertices1, vertices2, 0.0, 0.0)
        np.testing.assert_array_equal(assoc, expected_assoc)
        np.testing.assert_array_equal(common, expected_common)

    def test_association_condition1(self):
        """Check case where associations are made based on most common tracks."""
        vertices1 = np.array([
            [True, False, True, False, False, False, True],
            [False, True, False, True, False, False, False],
        ])
        vertices2 = np.array([
            [True, True, True, False, False, False, False],
        ])
        expected_assoc = np.array([
            [True],
            [False],
        ])
        expected_common = np.array([[2], [1]])
        assoc, common = associate_vertices(vertices1, vertices2, 0.0, 0.0)
        np.testing.assert_array_equal(assoc, expected_assoc)
        np.testing.assert_array_equal(common, expected_common)

    def test_association_condition2(self):
        """Check case where associations are made based on
        highest efficiency (tiebreaker 1).
        """
        vertices1 = np.array([
            [True, False, True, False, False, False, True],
            [False, True, False, True, False, False, False],
        ])
        vertices2 = np.array([
            [True, True, True, True, False, False, False],
        ])
        expected_assoc = np.array([
            [False],
            [True],
        ])
        expected_common = np.array([[2], [2]])
        assoc, common = associate_vertices(vertices1, vertices2, 0.0, 0.0)
        np.testing.assert_array_equal(assoc, expected_assoc)
        np.testing.assert_array_equal(common, expected_common)

    def test_association_condition3(self):
        """Check case where associations are made based on
        highest purity (tiebreaker 2).
        """
        vertices1 = np.array([
            [True, False, True, False, True, False, True],
        ])
        vertices2 = np.array([
            [True, False, True, False, False, False, False],
            [False, True, False, True, True, False, True],
        ])
        expected_assoc = np.array([[True, False]])
        expected_common = np.array([[2, 2]])
        assoc, common = associate_vertices(vertices1, vertices2, 0.0, 0.0)
        np.testing.assert_array_equal(assoc, expected_assoc)
        np.testing.assert_array_equal(common, expected_common)

    def test_association_condition4(self):
        """Check case where associations are made based on
        first match (tiebreaker 3).
        """
        vertices1 = np.array([
            [True, False, True, False, False, False, False],
            [False, True, False, True, False, False, False],
        ])
        vertices2 = np.array([
            [True, True, True, True, False, False, False],
        ])
        expected_assoc = np.array([
            [True],
            [False],
        ])
        expected_common = np.array([[2], [2]])
        assoc, common = associate_vertices(vertices1, vertices2, 0.0, 0.0)
        np.testing.assert_array_equal(assoc, expected_assoc)
        np.testing.assert_array_equal(common, expected_common)

    def test_full_associations(self):
        """Check case where associations are a perfect match."""
        vertices1 = np.array([
            [True, True, False, False, False],
        ])
        vertices2 = np.array([
            [True, True, False, False, False],
        ])
        expected_assoc = np.array([[True]])
        expected_common = np.array([[2]])
        assoc, common = associate_vertices(vertices1, vertices2, 0.0, 0.0)
        np.testing.assert_array_equal(assoc, expected_assoc)
        np.testing.assert_array_equal(common, expected_common)


class CalculateVertexMetricsTestCase(unittest.TestCase):
    """Test case for calculate_vertex_metrics function."""

    def test_no_vertices(self):
        """Check case where there are no vertices."""
        indices1 = np.array([[0, 1, 2, 3, 4]])
        indices2 = np.array([[0, 1, 2, 3, 4]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=2,
            eff_req=0.0,
            purity_req=0.0,
        )
        np.testing.assert_array_equal(metrics["n_match"], [0])
        np.testing.assert_array_equal(metrics["n_test"], [0])
        np.testing.assert_array_equal(metrics["n_ref"], [0])
        np.testing.assert_array_equal(metrics["track_overlap"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[-1, -1]])

    def test_no_reco_vertices(self):
        """Check case where there are no reco vertices."""
        indices1 = np.array([[0, 1, 2, 3, 4]])
        indices2 = np.array([[0, 1, 2, 2, 3]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=2,
            eff_req=0.0,
            purity_req=0.0,
        )
        np.testing.assert_array_equal(metrics["n_match"], [0])
        np.testing.assert_array_equal(metrics["n_test"], [0])
        np.testing.assert_array_equal(metrics["n_ref"], [1])
        np.testing.assert_array_equal(metrics["track_overlap"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[-1, -1]])

    def test_no_truth_vertices(self):
        """Check case where there are no truth vertices."""
        indices1 = np.array([[0, 0, 1, 2, 3]])
        indices2 = np.array([[0, 1, 2, 3, 4]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=2,
            eff_req=0.0,
            purity_req=0.0,
        )
        np.testing.assert_array_equal(metrics["n_match"], [0])
        np.testing.assert_array_equal(metrics["n_test"], [1])
        np.testing.assert_array_equal(metrics["n_ref"], [0])
        np.testing.assert_array_equal(metrics["track_overlap"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[-1, -1]])

    def test_no_match(self):
        """Check case where there are no vertex matches."""
        indices1 = np.array([[0, 0, 1, 2, 3]])
        indices2 = np.array([[0, 1, 2, 2, 3]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=2,
            eff_req=0.0,
            purity_req=0.0,
        )
        np.testing.assert_array_equal(metrics["n_match"], [0])
        np.testing.assert_array_equal(metrics["n_test"], [1])
        np.testing.assert_array_equal(metrics["n_ref"], [1])
        np.testing.assert_array_equal(metrics["track_overlap"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[-1, -1]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[-1, -1]])

    def test_one_match(self):
        """Check case where there is one vertex match."""
        indices1 = np.array([[0, 0, 1, 2, 2]])
        indices2 = np.array([[0, 0, 0, 1, 2]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=2,
            eff_req=0.0,
            purity_req=0.0,
        )
        np.testing.assert_array_equal(metrics["n_match"], [1])
        np.testing.assert_array_equal(metrics["n_test"], [2])
        np.testing.assert_array_equal(metrics["n_ref"], [1])
        np.testing.assert_array_equal(metrics["track_overlap"], [[2, -1]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[2, -1]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[3, -1]])

    def test_mult_matches(self):
        """Check case where there are multiple vertex matches."""
        indices1 = np.array([[0, 0, 1, 1, 0, 1, 2, 2, 3]])
        indices2 = np.array([[0, 1, 0, 2, 0, 2, 3, 3, 1]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=3,
            eff_req=0.0,
            purity_req=0.0,
        )
        np.testing.assert_array_equal(metrics["n_match"], [3])
        np.testing.assert_array_equal(metrics["n_test"], [3])
        np.testing.assert_array_equal(metrics["n_ref"], [4])
        np.testing.assert_array_equal(metrics["track_overlap"], [[2, 2, 2]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[3, 3, 2]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[3, 2, 2]])

    def test_eff_req(self):
        """Check case where efficiency requirement is not met."""
        indices1 = np.array([[0, 0, 1, 1, 2, 3]])
        indices2 = np.array([[0, 0, 1, 1, 1, 1]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=2,
            eff_req=0.7,
            purity_req=0.0,
        )
        np.testing.assert_array_equal(metrics["n_match"], [1])
        np.testing.assert_array_equal(metrics["n_test"], [2])
        np.testing.assert_array_equal(metrics["n_ref"], [2])
        np.testing.assert_array_equal(metrics["track_overlap"], [[2, -1]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[2, -1]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[2, -1]])

    def test_purity_req(self):
        """Check case where purity requirement is not met."""
        indices1 = np.array([[0, 0, 1, 1, 1, 1]])
        indices2 = np.array([[0, 0, 1, 1, 2, 3]])
        metrics = calculate_vertex_metrics(
            indices1,
            indices2,
            max_vertices=2,
            eff_req=0.0,
            purity_req=0.7,
        )
        np.testing.assert_array_equal(metrics["n_match"], [1])
        np.testing.assert_array_equal(metrics["n_test"], [2])
        np.testing.assert_array_equal(metrics["n_ref"], [2])
        np.testing.assert_array_equal(metrics["track_overlap"], [[2, -1]])
        np.testing.assert_array_equal(metrics["test_vertex_size"], [[2, -1]])
        np.testing.assert_array_equal(metrics["ref_vertex_size"], [[2, -1]])


class CleanTruthVerticesTestCase(unittest.TestCase):
    """Test case for clean_truth_vertices function."""

    def test_non_hf_vtx_removal(self):
        """Check case where non-pure HF vertices are removed."""
        vtx_indices = np.array([1, 0, 1, 0, 0, 1, 2, 0, 2, 2, 3, 3])
        trk_origin = np.array([2, 0, 2, 1, 1, 2, 2, 0, 4, 4, 3, 3])
        cleaned_indices = clean_truth_vertices(vtx_indices, trk_origin)
        expected_result = np.array([-99, -99, -99, -99, -99, -99, -99, -99, -99, -99, 3, 3])
        np.testing.assert_array_equal(cleaned_indices, expected_result)

    def test_incl_hf_vtx_merging(self):
        """Check case in inclusive vertexing where HF vertices are merged."""
        vtx_indices = np.array([0, 0, 1, 2, 1, 2, 2, 2])
        trk_origin = np.array([2, 2, 3, 4, 3, 4, 4, 4])
        cleaned_indices = clean_truth_vertices(vtx_indices, trk_origin, incl_vertexing=True)
        expected_result = np.array([-99, -99, 3, 3, 3, 3, 3, 3])
        np.testing.assert_array_equal(cleaned_indices, expected_result)


class CleanRecoVerticesTestCase(unittest.TestCase):
    """Test case for clean_reco_vertices function."""

    def test_pv_removal(self):
        """Check case where PV in model with track origin prediction is removed."""
        vtx_indices = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1])
        trk_origin = np.array([2, 4, 2, 2, 2, 2, 3, 3, 4])
        cleaned_indices = clean_reco_vertices(vtx_indices, trk_origin)
        expected_result = np.array([-99, -99, -99, -99, -99, 1, 1, 1, 1])
        np.testing.assert_array_equal(cleaned_indices, expected_result)

    def test_non_hf_vtx_removal(self):
        """Check case where vertices without HF track are removed."""
        vtx_indices = np.array([1, 0, 1, 0, 0, 1, 2, 0, 2, 2, 3, 3])
        trk_origin = np.array([2, 0, 2, 1, 1, 2, 2, 0, 4, 4, 3, 3])
        cleaned_indices = clean_reco_vertices(vtx_indices, trk_origin)
        expected_result = np.array([-99, -99, -99, -99, -99, -99, 2, -99, 2, 2, 3, 3])
        np.testing.assert_array_equal(cleaned_indices, expected_result)

    def test_incl_track_merging_hf(self):
        """Check case in incl vertexing w track origin prediction where HF vertices are merged."""
        vtx_indices = np.array([0, 0, 1, 2, 1, 2, 2, 1, 3, 3])
        trk_origin = np.array([2, 2, 3, 3, 3, 5, 7, 5, 7, 7])
        cleaned_indices = clean_reco_vertices(vtx_indices, trk_origin, incl_vertexing=True)
        expected_result = np.array([-99, -99, 3, 3, 3, 3, 3, 3, -99, -99])
        np.testing.assert_array_equal(cleaned_indices, expected_result)

    def test_incl_track_merging_all(self):
        """Check case in incl vertexing w/o trk_origin prediction where all vertices are merged."""
        vtx_indices = np.array([-2, -2, 0, 1, -2, 1, 1, 0])
        cleaned_indices = clean_reco_vertices(vtx_indices, incl_vertexing=True)
        expected_result = np.array([-2, -2, 2, 2, -2, 2, 2, 2])
        np.testing.assert_array_equal(cleaned_indices, expected_result)



from __future__ import annotations

import numpy as np


def get_fx_values(resolution=100):
    return np.concatenate((
        np.logspace(-3, -1, resolution // 2),
        np.linspace(0.1, 1.0, resolution // 2),
    ))


def get_efficiency(scores, fx):
    return np.sum(scores > fx) / len(scores)


def get_optimal_fc(fc_scan: np.ndarray, fc_space: np.ndarray, rej=False):
    """After calculating an fc scan, find the
    optimal value of fc.


    Parameters
    ----------
    fc_scan : np.ndarray
        2D array of efficiency (or rejection) scores
        for each value of fc.
    fc_space : np.ndarray
        1D array of fc values.
    rej : bool, optional
        If True, find the maximum rejection values else
        find the minimum efficiency values, by default False.
    """
    # normalise x- and y-axes
    xs, ys = fc_scan[:, 0], fc_scan[:, 1]
    xs = xs / max(xs)
    ys = ys / max(ys)

    # if rej=True get maximum distance to origin
    opt_idx = np.argmax(xs**2 + ys**2) if rej else np.argmin(xs**2 + ys**2)

    return opt_idx, fc_space[opt_idx]
